
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Gui } from "Framework/Gui/Gui";
import { Graph } from "Framework/Gui/Graph";
import { TimeMeas } from "Framework/Time/TimeMeas";

import { LaplaceTransformFunction } from "./LaplaceTransformFunction";

import { ILaplaceTransformInversion } from "Framework/LaplaceTransform/ILaplaceTransformInversion";
import { LaplaceTransformInversionHuddleston } from "Framework/LaplaceTransform/LaplaceTransformInversionHuddleston";
import { LaplaceTransformInversionKoizumi } from "Framework/LaplaceTransform/LaplaceTransformInversionKoizumi";
import { LaplaceTransformInversionStehfest } from "Framework/LaplaceTransform/LaplaceTransformInversionStehfest";
import { LaplaceTransformInversionZakian } from "Framework/LaplaceTransform/LaplaceTransformInversionZakian";
import { LaplaceTransformInversionTalbot } from "Framework/LaplaceTransform/LaplaceTransformInversionTalbot";

import { ControlElement } from "../../shared/ControlElement";

//---------------------------------------------------------------------------------------

declare var MathJax: any;

//-------------------------------------------------------------------------------------------------

export class ControlAppLaplaceTransform
{
    public static Inst: ControlAppLaplaceTransform = new ControlAppLaplaceTransform();

    //-------------------------------------------------------------------------------------------------

    public Init(): void
    {
        this.FunctionSelected();
    }

    //-------------------------------------------------------------------------------------------------

    public FunctionSelected(): void
    {
        let funcName: string = Gui.GetComboboxItem("FUNC");
        let htmlElemTransferFunc: HTMLElement = document.getElementById("TRANSFER_FUNC");
        let htmlElemStepResponse: HTMLElement = document.getElementById("STEP_RESPONSE");

        htmlElemTransferFunc.textContent = "$$" + ControlElement.GetTransferFunctionLatex(funcName) + "$$";
        MathJax.Hub.Queue(["Typeset", MathJax.Hub, htmlElemTransferFunc]);

        htmlElemStepResponse.textContent = "$$" + ControlElement.GetStepResponseLatex(funcName) + "$$";
        MathJax.Hub.Queue(["Typeset", MathJax.Hub, htmlElemStepResponse]);
    }

    //-------------------------------------------------------------------------------------------------

    public NumericLaplaceInversion(): void
    {
        try
        {
            let funcName: string = Gui.GetComboboxItem("FUNC");
            let method1: string = Gui.GetComboboxItem("METHOD_1");
            let method2: string = Gui.GetComboboxItem("METHOD_2");
            let endTime: number = Gui.GetNumber("END_TIME");
            let timeStep: number = Gui.GetNumber("TIME_STEP");

            let numPoints: number = 1000;
            let idx: number;

            let func: LaplaceTransformFunction = this.GetFunction(funcName);
            let curveAna: [number, number][] = [];

            for (idx = 0; idx < numPoints; idx++)
            {
                let t: number = endTime / (numPoints - 1) * idx;
                curveAna[idx] = [t, func.TimeFunctionStepResponse(t)];
            }

            let curveNameList: string[] = ["analytical"];
            let curveAxisList: number[] = [1];
            let curveColorList: string[] = ["#EE0000"];
            let curveWidthList: number[] = [4];
            let curveList: [number, number][][] = [curveAna];
            let timeMeas: TimeMeas = new TimeMeas();

            if (method1.toLowerCase() != "none")
            {
                curveNameList.push(method1);
                curveAxisList.push(1);
                curveColorList.push("#00EE00");
                curveWidthList.push(1);

                timeMeas.Reset();
                let curve: [number, number][] = this.GetCurve(method1, func, endTime, timeStep, numPoints);
                let time: number = timeMeas.Measure();

                curveList.push(curve);

                Gui.SetTableCell("TIME", 0, 0, "Time " + method1)
                Gui.SetTableCell("TIME", 1, 0, (time * 1000).toPrecision(4) + " ms")
            }
            else
            {
                Gui.SetTableCell("TIME", 0, 0, "")
                Gui.SetTableCell("TIME", 1, 0, "")
            }

            if (method2.toLowerCase() != "none")
            {
                curveNameList.push(method2);
                curveAxisList.push(1);
                curveColorList.push("#0000EE");
                curveWidthList.push(1);

                timeMeas.Reset();
                let curve: [number, number][] = this.GetCurve(method2, func, endTime, timeStep, numPoints);
                let time: number = timeMeas.Measure();

                curveList.push(curve);

                Gui.SetTableCell("TIME", 0, 1, "Time " + method2)
                Gui.SetTableCell("TIME", 1, 1, (time * 1000).toPrecision(4) + " ms")
            }
            else
            {
                Gui.SetTableCell("TIME", 0, 1, "")
                Gui.SetTableCell("TIME", 1, 1, "")
            }

            Graph.Draw("DIV_GRAPH", "Step response", "Time", "Amplitude", "", curveNameList, curveAxisList, curveColorList, curveList, curveWidthList);
        }
        catch (err)
        {
            alert(err.message);
        }
    }

    //-------------------------------------------------------------------------------------------------

    private GetFunction(func: string): LaplaceTransformFunction
    {
        let K: number = Gui.GetNumber("PARAM_K");
        let T: number = Gui.GetNumber("PARAM_T");
        let D: number = Gui.GetNumber("PARAM_D");

        return LaplaceTransformFunction.Create(func, K, T, D);
    }

    //-------------------------------------------------------------------------------------------------

    private GetCurve(methodName: string, func: LaplaceTransformFunction, endTime: number, timeStep: number, numPoints: number)
    {
        if (methodName.toLocaleLowerCase() == "deq")
        {
            return func.TimeFunctionDifferentialEquation(endTime, timeStep, numPoints);
        }
        else
        {
            let ilt: ILaplaceTransformInversion;

            switch (methodName.toLowerCase())
            {
                case "huddleston": ilt = new LaplaceTransformInversionHuddleston(); break;
                case "koizumi": ilt = new LaplaceTransformInversionKoizumi(); break;
                case "stehfest": ilt = new LaplaceTransformInversionStehfest(); break;
                case "zakian": ilt = new LaplaceTransformInversionZakian(); break;
                case "talbot": ilt = new LaplaceTransformInversionTalbot(); break;
                default: throw new Error("invalid method: " + methodName);
            }

            let idx: number;
            let tArr: number[] = [];

            for (idx = 0; idx < numPoints; idx++)
                tArr[idx] = endTime / (numPoints - 1) * idx;

            let yArr: number[] = ilt.Calculate(func, tArr);
            let curve: [number, number][] = [];

            for (idx = 0; idx < numPoints; idx++)
                curve[idx] = [tArr[idx], yArr[idx]];

            return curve;
        }
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------

