
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "Framework/Math/Complex";
import { IComplexFunction } from "Framework/Math/IComplexFunction";

import { ControlElement } from "../../shared/ControlElement";
import { ControlElementD } from "../../shared/ControlElementD";
import { ControlElementDeadTime } from "../../shared/ControlElementDeadTime";
import { ControlElementI } from "../../shared/ControlElementI";
import { ControlElementP } from "../../shared/ControlElementP";
import { ControlElementPD1 } from "../../shared/ControlElementPD1";
import { ControlElementPT1 } from "../../shared/ControlElementPT1";
import { ControlElementPT2 } from "../../shared/ControlElementPT2";

//-------------------------------------------------------------------------------------------------

export class LaplaceTransformFunction implements IComplexFunction
{
    private _ctrlElem: ControlElement;

    private constructor(ctrlElem: ControlElement)
    {
        this._ctrlElem = ctrlElem;
    }

    public static Create(type: string, K?: number, T?: number, D?: number): LaplaceTransformFunction
    {
        switch (type)
        {
            case "D": return new LaplaceTransformFunction(new ControlElementD(K));
            case "DeadTime": return new LaplaceTransformFunction(new ControlElementDeadTime(T));
            case "I": return new LaplaceTransformFunction(new ControlElementI(T));
            case "P": return new LaplaceTransformFunction(new ControlElementP(K));
            case "PD1": return new LaplaceTransformFunction(new ControlElementPD1(K, T));
            case "PT1": return new LaplaceTransformFunction(new ControlElementPT1(K, T));
            case "PT2": return new LaplaceTransformFunction(new ControlElementPT2(K, T, D));
            default: throw new Error("unknown type: " + type);
        }
    }

    public TimeFunctionStepResponse(t: number): number
    {
        return this._ctrlElem.TimeFunctionStepResponse(t);
    }

    public TransferFunction(s: Complex): Complex
    {
        return this._ctrlElem.TransferFunction(s);
    }

    private _res: Complex = new Complex();

    public Function(s: Complex): Complex
    {
        this._res.Copy(this.TransferFunction(s));
        this._res.Div(s);
        return this._res;
    }

    public TimeFunctionDifferentialEquation(endTime: number, timeStep: number, numPoints: number): [number, number][]
    {
        let u: number = 1;
        let yArr: [number, number][] = [];
        let t: number = 0;
        let idx: number = 0;
        let dt: number = endTime / (numPoints - 1);
        let tt: number = dt;

        this._ctrlElem.Initialize(timeStep);

        for (t = 0; t <= endTime; t += timeStep)
        {
            let y: number = this._ctrlElem.Calculate(u, timeStep);

            tt += timeStep;
            if (tt >= dt)
            {
                tt -= dt;
                yArr[idx] = [t, y];
                idx += 1;
            }
        }

        return yArr;
    }
}

//-------------------------------------------------------------------------------------------------

