
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Gui } from "Framework/Gui/Gui";

import { ControlAppLoopStepResponse } from "../../ControlAppLoopStepResponse/src/ControlAppLoopStepResponse";
import { ControlPath } from "../../shared/ControlPath";
import { ControlLoop } from "../../shared/ControlLoop";
import { ControlPID } from "../../shared/ControlPID";
import { ControllerOptimizationLocalVariation } from "../../shared/ControllerOptimizationLocalVariation";
import { ControllerOptimizationDifferentialEvolution } from "../../shared/ControllerOptimizationDifferentialEvolution";

//-------------------------------------------------------------------------------------------------

export class ControlAppLoopTuningIteration extends ControlAppLoopStepResponse
{
    public static Inst: ControlAppLoopTuningIteration = new ControlAppLoopTuningIteration();

    //-------------------------------------------------------------------------------------------------

    public Init(): void
    {
        this.ControlAppInit();
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningLocalVariation(): void
    {
        try
        {
            let pidKp: number = Gui.GetNumber("PID_KP");
            let pidKi: number = Gui.GetNumber("PID_KI");
            let pidKd: number = Gui.GetNumber("PID_KD");
            let pidUMin: number = Gui.GetNumber("PID_U_MIN");
            let pidUMax: number = Gui.GetNumber("PID_U_MAX");

            let stepHeight: number = Gui.GetNumber("STEP_HEIGHT");
            let stepTime: number = Gui.GetNumber("STEP_TIME");
            let endTime: number = Gui.GetNumber("END_TIME");
            let maxNumSteps: number = Gui.GetNumber("MAX_NUM_STEPS");
            let maxOvershotPercent: number = Gui.GetNumber("MAX_OVERSHOT");

            let weightingCtrlOut: number = Gui.GetNumber("WEIGHTING_CTRL_OUT");
            let weightingOvershot: number = Gui.GetNumber("WEIGHTING_OVERSHOT");

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlPid: ControlPID = new ControlPID(pidKp, pidKi, pidKd, pidUMin, pidUMax);
            let ctrlLoop: ControlLoop = new ControlLoop(ctrlPid, ctrlPath);
            let ctrlOpti: ControllerOptimizationLocalVariation = new ControllerOptimizationLocalVariation(ctrlLoop, ctrlPid);

            let startParams: number[] = [pidKp, pidKi, pidKd];
            let startDelta: number[] = [pidKp + 10, pidKi + 10, pidKd + 10];
            let maxOvershot: number = Math.abs(maxOvershotPercent) / 100 * stepHeight;

            let params: number[] = ctrlOpti.Optimize(startParams, startDelta, maxNumSteps,
                stepHeight, stepTime, endTime, maxOvershot,
                weightingCtrlOut, weightingOvershot);

            Gui.Set("PID_KP", params[0]);
            Gui.Set("PID_KI", params[1]);
            Gui.Set("PID_KD", params[2]);

            this.SetControlParamInGuiPic("PID_CTRL_1_KP", "Kp = " + params[0].toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KI", "Ki = " + params[1].toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KD", "Kd = " + params[2].toPrecision(3));
        }
        catch (err)
        {
            alert(err.message);
        }

        this.ControlLoopStepResponse();
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningDiffEvo(): void
    {
        try
        {
            let pidKp: number = Gui.GetNumber("PID_KP");
            let pidKi: number = Gui.GetNumber("PID_KI");
            let pidKd: number = Gui.GetNumber("PID_KD");
            let pidUMin: number = Gui.GetNumber("PID_U_MIN");
            let pidUMax: number = Gui.GetNumber("PID_U_MAX");

            let stepHeight: number = Gui.GetNumber("STEP_HEIGHT");
            let stepTime: number = Gui.GetNumber("STEP_TIME");
            let endTime: number = Gui.GetNumber("END_TIME");
            let maxNumSteps: number = Gui.GetNumber("MAX_NUM_STEPS");
            let maxOvershotPercent: number = Gui.GetNumber("MAX_OVERSHOT");

            let weightingCtrlOut: number = Gui.GetNumber("WEIGHTING_CTRL_OUT");
            let weightingOvershot: number = Gui.GetNumber("WEIGHTING_OVERSHOT");

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlPid: ControlPID = new ControlPID(pidKp, pidKi, pidKd, pidUMin, pidUMax);
            let ctrlLoop: ControlLoop = new ControlLoop(ctrlPid, ctrlPath);
            let ctrlOpti: ControllerOptimizationDifferentialEvolution = new ControllerOptimizationDifferentialEvolution(ctrlLoop, ctrlPid);

            let maxOvershot: number = Math.abs(maxOvershotPercent) / 100 * stepHeight;
            let dither: boolean = true;

            let params: number[] = ctrlOpti.Optimize(0, 100, dither, maxNumSteps,
                stepHeight, stepTime, endTime, maxOvershot,
                weightingCtrlOut, weightingOvershot);

            Gui.Set("PID_KP", params[0]);
            Gui.Set("PID_KI", params[1]);
            Gui.Set("PID_KD", params[2]);

            this.SetControlParamInGuiPic("PID_CTRL_1_KP", "Kp = " + params[0].toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KI", "Ki = " + params[1].toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KD", "Kd = " + params[2].toPrecision(3));
        }
        catch (err)
        {
            alert(err.message);
        }

        this.ControlLoopStepResponse();
    }

    //-------------------------------------------------------------------------------------------------
}
