
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Gui } from "Framework/Gui/Gui";
import { Graph } from "Framework/Gui/Graph";

import { ControlApp } from "../../shared/ControlApp";
import { ControlTools } from "../../shared/ControlTools";
import { ControlPath } from "../../shared/ControlPath";
import { ControlLoop } from "../../shared/ControlLoop";
import { ControlSimulation } from "../../shared/ControlSimulation";
import { ControlPID } from "../../shared/ControlPID";
import { ControlComparator } from "../../shared/ControlComparator";

//-------------------------------------------------------------------------------------------------

export class ControlAppLoopTuningOscillation extends ControlApp
{
    public static Inst: ControlAppLoopTuningOscillation = new ControlAppLoopTuningOscillation();

    //-------------------------------------------------------------------------------------------------

    public Init(): void
    {
        this.ControlAppInit();
    }

    //-------------------------------------------------------------------------------------------------

    private _ku: number = 0;
    private _pu: number = 0;

    public ControlLoopTuningOscillation(): void
    {
        try
        {
            let comparatorTh: number = Gui.GetNumber("COMPARATOR_TH");
            let comparatorUMin: number = Gui.GetNumber("COMPARATOR_U_MIN");
            let comparatorUMax: number = Gui.GetNumber("COMPARATOR_U_MAX");
            let endTime: number = Gui.GetNumber("END_TIME");

            this.SetControlParamInGuiPic("COMPARATOR_TH", "Th = " + comparatorTh);

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlComparator: ControlComparator = new ControlComparator(comparatorTh, comparatorUMin, comparatorUMax);
            let ctrlLoop: ControlLoop = new ControlLoop(ctrlComparator, ctrlPath);
            let ctrlSim: ControlSimulation = new ControlSimulation(ctrlLoop);
            let curveW: [number, number][];
            let curveU: [number, number][];
            let curveY: [number, number][];

            [curveW, curveU, curveY] = ctrlSim.StepResponseControlLoop(0, 0, endTime);

            let au: number = ctrlComparator.Amplitude;
            let ay: number = ctrlLoop.Amplitude;
            let p: number = ctrlComparator.Period;

            this._pu = p;
            this._ku = 4 * au / (Math.PI * ay);

            Gui.SetTableCell("TABLE_PATH_PARAM", 1, 2, au.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 2, 2, ay.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 3, 2, p.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 4, 2, this._ku.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 5, 2, this._pu.toPrecision(4));

            let kp: number;
            let ki: number;
            let kd: number;

            [kp, ki, kd] = ControlTools.GetPidParamsZieglerNichols(this._pu, this._ku);
            this.ShowPidParams(1, kp, ki, kd);

            [kp, ki, kd] = ControlTools.GetPidParamsLittleOvershot(this._pu, this._ku);
            this.ShowPidParams(2, kp, ki, kd);

            [kp, ki, kd] = ControlTools.GetPidParamsNoOvershot(this._pu, this._ku);
            this.ShowPidParams(3, kp, ki, kd);

            Graph.Draw("DIV_GRAPH", "PID tuning relay method", "Time", "Amplitude", "Amplitude",
                ["u(t)", "y(t)"], [1, 1], ["#0000FF", "#FF0000"], [curveU, curveY]);

            this.ControlLoopTuningOscillationParamMethodSelected();
        }
        catch (err)
        {
            alert(err.message);
        }
    }

    //-------------------------------------------------------------------------------------------------

    private ShowPidParams(row: number, kp: number, ki: number, kd: number): void
    {
        Gui.SetTableCell("TABLE_PID_PARAMS", row, 1, kp.toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAMS", row, 3, ki.toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAMS", row, 2, (kp / ki).toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAMS", row, 5, kd.toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAMS", row, 4, (kd / kp).toPrecision(4));
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningOscillationParamMethodSelected(): void
    {
        let kp: number;
        let ki: number;
        let kd: number;
        let pidParams: string = Gui.GetString("PID_PARAMS");

        switch (pidParams)
        {
            case "L":
                [kp, ki, kd] = ControlTools.GetPidParamsLittleOvershot(this._pu, this._ku);
                break;
            case "N":
                [kp, ki, kd] = ControlTools.GetPidParamsNoOvershot(this._pu, this._ku);
                break;
            case "Z":
            default:
                [kp, ki, kd] = ControlTools.GetPidParamsZieglerNichols(this._pu, this._ku);
                break;
        }

        Gui.Set("PID_KP", kp);
        Gui.Set("PID_KI", ki);
        Gui.Set("PID_KD", kd);
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningOscillationStepResponse(): void
    {
        try
        {
            let pidKp: number = Gui.GetNumber("PID_KP");
            let pidKi: number = Gui.GetNumber("PID_KI");
            let pidKd: number = Gui.GetNumber("PID_KD");
            let pidUMin: number = Gui.GetNumber("PID_U_MIN");
            let pidUMax: number = Gui.GetNumber("PID_U_MAX");
            let stepHeight: number = Gui.GetNumber("STEP_HEIGHT");
            let stepTime: number = Gui.GetNumber("STEP_TIME");
            let endTime: number = Gui.GetNumber("END_TIME_STEP_RESPONSE");

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlPid: ControlPID = new ControlPID(pidKp, pidKi, pidKd, pidUMin, pidUMax);
            let ctrlLoop: ControlLoop = new ControlLoop(ctrlPid, ctrlPath);
            let ctrlSim: ControlSimulation = new ControlSimulation(ctrlLoop);
            let curveW: [number, number][];
            let curveU: [number, number][];
            let curveY: [number, number][];
            let maxY: number;

            [curveW, curveU, curveY, maxY] = ctrlSim.StepResponseControlLoop(stepHeight, stepTime, endTime);

            Graph.Draw("DIV_GRAPH_STEP_RESPONSE", "Step response", "Time", "Amplitude", "Amplitude",
                ["u(t)", "w(t)", "y(t)"], [2, 1, 1], ["#00FF00", "#0000FF", "#FF0000"], [curveU, curveW, curveY]);

            let overshotPercent: number = (maxY - stepHeight) / stepHeight * 100;
            document.getElementById("OVERSHOT").textContent = "Überschwingen: " + overshotPercent.toFixed(2) + "%";
        }
        catch (err)
        {
            alert(err.message);
        }
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
