
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Gui } from "Framework/Gui/Gui";
import { Graph } from "Framework/Gui/Graph";

import { ControlApp } from "../../shared/ControlApp";
import { ControlTools } from "../../shared/ControlTools";
import { ControlPath } from "../../shared/ControlPath";
import { ControlLoop } from "../../shared/ControlLoop";
import { ControlSimulation } from "../../shared/ControlSimulation";
import { ControlPID } from "../../shared/ControlPID";

//-------------------------------------------------------------------------------------------------

export class ControlAppLoopTuningStepResponse extends ControlApp
{
    public static Inst: ControlAppLoopTuningStepResponse = new ControlAppLoopTuningStepResponse();

    //-------------------------------------------------------------------------------------------------

    public Init(): void
    {
        this.ControlAppInit();
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningStepResponse(graphId: string): void
    {
        const X: number = 0;
        const Y: number = 1;

        try
        {
            let stepHeight: number = Gui.GetNumber("STEP_HEIGHT");
            let stepTime: number = Gui.GetNumber("STEP_TIME");
            let endTime: number = Gui.GetNumber("END_TIME");

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlSim: ControlSimulation = new ControlSimulation(ctrlPath);
            let curveU: [number, number][];
            let curveY: [number, number][];

            [curveU, curveY] = ctrlSim.StepResponseControlPath(stepHeight, stepTime, endTime);

            let m: number;
            let n: number;

            [m, n] = ControlTools.GetInflectionTangent(curveY, stepTime);

            let y0: number = 0;
            let x0: number = -n / m;
            let ye: number = curveY[curveY.length - 1][Y];
            let xe: number = (ye - n) / m;

            let curveT: [number, number][] = [[x0, y0], [xe, ye]];
            let curveTg: [number, number][] = [[xe, 0], [xe, ye]];
            let curveTu: [number, number][] = [[x0, 0], [x0, ye]];
            let curveDX: [number, number][] = [[0, ye], [curveY[curveY.length - 1][X], ye]];

            Graph.Draw(graphId, "Step response", "Time", "Amplitude", "Amplitude",
                ["u(t)", "y(t)", "tangent", "Tg", "Tu", "dX"],
                [1, 1, 1, 1, 1, 1],
                ["#0000FF", "#FF0000", "#00FF00", "#555555", "#777777", "#999999"],
                [curveU, curveY, curveT, curveTg, curveTu, curveDX]);

            let ks: number = ye - y0;
            let tu: number = x0 - stepTime;
            let tg: number = xe - x0;

            Gui.SetTableCell("TABLE_PATH_PARAM", 1, 2, ks.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 2, 2, tu.toPrecision(4));
            Gui.SetTableCell("TABLE_PATH_PARAM", 3, 2, tg.toPrecision(4));

            this.CalcPidParamTable(ks, tu, tg);
        }
        catch (err)
        {
            alert(err.message);
        }
    }

    //-------------------------------------------------------------------------------------------------

    private _ctrlLoopStepTuningPidVals: { [key: string]: [number, number, number] } = {};

    private CalcPidParamTable(ks: number, tu: number, tg: number): void
    {
        this._ctrlLoopStepTuningPidVals["ZN_P"] = [1.0 * tg / (ks * tu), 0, 0];
        this._ctrlLoopStepTuningPidVals["ZN_PI"] = [0.9 * tg / (ks * tu), 0.273 * tg / (ks * tu * tu), 0];
        this._ctrlLoopStepTuningPidVals["ZN_PID"] = [1.2 * tg / (ks * tu), 0.6 * tg / (ks * tu * tu), 0.6 * tg / ks];
        this._ctrlLoopStepTuningPidVals["CHR_P"] = [0.3 * tg / (ks * tu), 0, 0];
        this._ctrlLoopStepTuningPidVals["CHR_PI"] = [0.35 * tg / (ks * tu), 0.292 / (ks * tu), 0];
        this._ctrlLoopStepTuningPidVals["CHR_PID"] = [0.6 * tg / (ks * tu), 0.6 / (ks * tu), 0.3 * tg / ks];

        Gui.SetTableCell("TABLE_PID_PARAM", 1, 1, this._ctrlLoopStepTuningPidVals["ZN_P"][0].toPrecision(4));

        Gui.SetTableCell("TABLE_PID_PARAM", 2, 1, this._ctrlLoopStepTuningPidVals["ZN_PI"][0].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 2, 2, this._ctrlLoopStepTuningPidVals["ZN_PI"][1].toPrecision(4));

        Gui.SetTableCell("TABLE_PID_PARAM", 3, 1, this._ctrlLoopStepTuningPidVals["ZN_PID"][0].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 3, 2, this._ctrlLoopStepTuningPidVals["ZN_PID"][1].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 3, 3, this._ctrlLoopStepTuningPidVals["ZN_PID"][2].toPrecision(4));

        Gui.SetTableCell("TABLE_PID_PARAM", 4, 1, this._ctrlLoopStepTuningPidVals["CHR_P"][0].toPrecision(4));

        Gui.SetTableCell("TABLE_PID_PARAM", 5, 1, this._ctrlLoopStepTuningPidVals["CHR_PI"][0].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 5, 2, this._ctrlLoopStepTuningPidVals["CHR_PI"][1].toPrecision(4));

        Gui.SetTableCell("TABLE_PID_PARAM", 6, 1, this._ctrlLoopStepTuningPidVals["CHR_PID"][0].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 6, 2, this._ctrlLoopStepTuningPidVals["CHR_PID"][1].toPrecision(4));
        Gui.SetTableCell("TABLE_PID_PARAM", 6, 3, this._ctrlLoopStepTuningPidVals["CHR_PID"][2].toPrecision(4));
    }

    //-------------------------------------------------------------------------------------------------

    public ControlLoopTuningDrawStepResponse(graphId: string): void
    {
        try
        {
            let method: string = Gui.GetCheckedRadioValue("PID_VALS");

            let pidKp: number = this._ctrlLoopStepTuningPidVals[method][0];
            let pidKi: number = this._ctrlLoopStepTuningPidVals[method][1];
            let pidKd: number = this._ctrlLoopStepTuningPidVals[method][2];

            let pidUMin: number = Gui.GetNumber("PID_U_MIN");
            let pidUMax: number = Gui.GetNumber("PID_U_MAX");

            let stepHeight: number = Gui.GetNumber("STEP_HEIGHT");
            let stepTime: number = Gui.GetNumber("STEP_TIME");
            let endTime: number = Gui.GetNumber("END_TIME");

            this.SetControlParamInGuiPic("PID_CTRL_1_KP", "Kp = " + pidKp.toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KI", "Ki = " + pidKi.toPrecision(3));
            this.SetControlParamInGuiPic("PID_CTRL_1_KD", "Kd = " + pidKd.toPrecision(3));

            let ctrlPath: ControlPath = this.CreateControlPathFromGui();
            let ctrlPid: ControlPID = new ControlPID(pidKp, pidKi, pidKd, pidUMin, pidUMax);
            let ctrlLoop: ControlLoop = new ControlLoop(ctrlPid, ctrlPath);
            let ctrlSim: ControlSimulation = new ControlSimulation(ctrlLoop);
            let curveW: [number, number][];
            let curveU: [number, number][];
            let curveY: [number, number][];
            let maxY: number;

            [curveW, curveU, curveY, maxY] = ctrlSim.StepResponseControlLoop(stepHeight, stepTime, endTime);

            let curveNameList: string[] = ["u(t)", "w(t)", "y(t)"];
            let yAxisList: number[] = [2, 1, 1];
            let curveColorList: string[] = ["#00FF00", "#0000FF", "#FF0000"];
            let curveDataList: [number, number][][] = [curveU, curveW, curveY];

            Graph.Draw(graphId, "Step response", "Time", "Amplitude", "Amplitude",
                curveNameList, yAxisList, curveColorList, curveDataList);

            let overshotPercent: number = (maxY - stepHeight) / stepHeight * 100;
            document.getElementById("OVERSHOT").textContent = "Überschwingen: " + overshotPercent.toFixed(2) + "%";
        }
        catch (err)
        {
            alert(err.message);
        }
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
