
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { IControlOptimization } from "./IControlOptimization";
import { Complex } from "../../Framework/Math/Complex";

//-------------------------------------------------------------------------------------------------

export abstract class ControlElement implements IControlOptimization
{
    public abstract GetType(): string;
    public abstract GetLowestTimeConstant(): number;

    public abstract Initialize(dt: number): void;
    public abstract Calculate(input: number, dt: number): number;

    public abstract OptimizeGetNumParameters(): number;
    public abstract OptimizeSetParameters(params: number[], startIdx: number): void;

    public abstract TimeFunctionStepResponse(t: number): number;
    public abstract TransferFunction(s: Complex): Complex;

    public static GetTransferFunctionLatex(func: string): string
    {
        if (func in this.TRANSFER_FUNCTION)
            return this.TRANSFER_FUNCTION[func];
        else
            return "";
    }

    public static GetStepResponseLatex(func: string): string
    {
        if (func in this.STEP_RESPONSE)
            return this.STEP_RESPONSE[func];
        else
            return "";
    }

    private static readonly TRANSFER_FUNCTION: { [key: string]: string } = {
        "P": "K",
        "I": "\\frac{1}{T \\cdot s}",
        "D": "T \\cdot s",
        "PT1": "\\frac{K}{1 + T \\cdot s}",
        "PT2": "\\frac{K}{1 + 2 \\cdot D \\cdot T \\cdot s + T^2 \\cdot s^2}",
        "PD1": "K \\cdot (1 + T \\cdot s)",
        "DeadTime": "e^{-s \\cdot T}",
        "PID": "K_P+\\frac{K_I}{s}+K_D \\cdot s"
    };

    private static readonly STEP_RESPONSE: { [key: string]: string } = {
        "P": "f(t) = K",
        "I": "f(t) = \\frac{t}{T}",
        "D": "f(t) = T \\cdot \\delta (t)",
        "PT1": "f(t) = K (1 - e^{\\frac{-t}{T}})",
        "PD1": "f(t) = K ( T \\cdot \\delta(t)+1)",
        "DeadTime": "f(t) = \\Theta (t - T)",
        "PID": "f(t) = K_P + K_I \\cdot t + K_D \\cdot \\delta (t)",
        "PT2": "\\begin{align} A &= (D^2-1)T^2 \\\\ f(t) &= K(1-\\frac{\\sqrt{A}}{2 A}((\\sqrt{A}-D T) e^{t(-\\frac{\\sqrt{A}}{T^2}-\\frac{D}{T})} + (\\sqrt{A} + D T) e^{t(\\frac{\\sqrt{A}}{T^2}-\\frac{D}{T})} )) \\end{align}"
    };
}

//-------------------------------------------------------------------------------------------------
