
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { Complex } from "../../Framework/Math/Complex";

//-------------------------------------------------------------------------------------------------

export class ControlElementPT2 extends ControlElement
{
    private _K: number;
    private _T: number;
    private _D: number;

    private _y_n1: number = 0;
    private _y_n2: number = 0;

    public constructor(K: number, T: number, D: number)
    {
        super();

        if (T < 0)
            throw new Error("T must be greater then zero!");

        if (D < 0)
            throw new Error("D must be greater then zero!");

        this._K = K;
        this._T = T;
        this._D = D;
    }

    public GetType(): string
    {
        return "PT2";
    }

    public GetLowestTimeConstant(): number
    {
        return this._T;
    }

    public Initialize(_dt: number): void
    {
        this._y_n1 = 0;
        this._y_n2 = 0;
    }

    public Calculate(u: number, dt: number): number
    {
        let K: number = this._K;
        let T: number = this._T;
        let D: number = this._D;
        let y_n2: number = this._y_n2;
        let y_n1: number = this._y_n1;

        let y: number = ((2 * y_n1 - y_n2) * T * T + 2 * dt * y_n1 * D * T + dt * dt * u * K) / (T * T + 2 * dt * D * T + dt * dt);

        this._y_n2 = this._y_n1;
        this._y_n1 = y;

        return y;
    }

    public TimeFunctionStepResponse(t: number): number
    {
        // a = (D^2 - 1) * T
        let a: number = (this._D * this._D - 1) * this._T * this._T;

        // b = Sqrt(a)
        let b: Complex = new Complex(a, 0);
        b.Sqrt();

        // c = Exp(t * (-b / T^2 - D / T))
        let c: Complex = new Complex();
        c.Copy(b);
        c.Neg();
        c.DivR(this._T * this._T);
        c.SubR(this._D / this._T);
        c.MulR(t);
        c.Exp();

        // d = Exp(t * (b / T^2 - D / T))
        let d: Complex = new Complex();
        d.Copy(b);
        d.DivR(this._T * this._T);
        d.SubR(this._D / this._T);
        d.MulR(t);
        d.Exp();

        let e: Complex = new Complex();
        let f: Complex = new Complex();

        f.Copy(c);
        f.MulR(this._D * this._T);
        e.Sub(f);

        f.Copy(d);
        f.MulR(this._D * this._T);
        e.Add(f);

        f.Copy(c);
        f.Mul(b);
        e.Add(f);

        f.Copy(d);
        f.Mul(b);
        e.Add(f);

        e.Mul(b);
        e.MulR(-1 / (2 * a));
        e.AddR(1);

        e.MulR(this._K);

        return e.Real;
    }

    private _res: Complex = new Complex(0);

    public TransferFunction(s: Complex): Complex
    {
        // K / (1 + (2 * D * T + T * T * s) * s)

        this._res.Copy(s);
        this._res.MulR(this._T * this._T);
        this._res.AddR(2 * this._D * this._T);
        this._res.Mul(s);
        this._res.AddR(1);
        this._res.DivR(this._K);
        this._res.Inverse();

        return this._res;
    }

    public OptimizeGetNumParameters(): number
    {
        return 3;
    }

    public OptimizeSetParameters(params: number[], startIdx: number): void
    {
        this._K = params[startIdx + 0];
        this._T = params[startIdx + 1];
        this._D = params[startIdx + 2];

        if (this._T < 0)
            this._T = 0;

        if (this._D < 0)
            this._D = 0;
    }
}

//-------------------------------------------------------------------------------------------------
