
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { Complex } from "../../Framework/Math/Complex";

//-------------------------------------------------------------------------------------------------

export class ControlLoop extends ControlElement
{
    private readonly _controller: ControlElement;
    private readonly _controlPath: ControlElement;

    private _u: number = 0;
    private _y_n1: number = 0;
    //private _y_n2: number = 0;

    private _yMin: number = 0;
    private _yMax: number = 0;

    private _direction: number;

    public get U(): number { return this._u; }
    public get YMin(): number { return this._yMin; }
    public get YMax(): number { return this._yMax; }

    public get Amplitude(): number { return (this._yMax - this._yMin) / 2; }

    public constructor(controller: ControlElement, controlPath: ControlElement)
    {
        super();

        this._controller = controller;
        this._controlPath = controlPath;
    }

    public GetType(): string
    {
        return "ControlLoop";
    }

    public GetLowestTimeConstant(): number
    {
        let t1: number = this._controller.GetLowestTimeConstant();
        let t2: number = this._controlPath.GetLowestTimeConstant();

        if (t1 == -1)
            return t2;

        if (t2 == -1)
            return t1;

        return Math.min(t1, t2);
    }

    public Initialize(dt: number): void
    {
        this._u = 0;
        this._y_n1 = 0;
        //this._y_n2 = 0;

        this._yMin = 0;
        this._yMax = 0;

        this._direction = 1;

        this._controller.Initialize(dt);
        this._controlPath.Initialize(dt);
    }

    public Calculate(w: number, dt: number): number
    {
        let e: number = w - this._y_n1;
        let u: number = this._controller.Calculate(e, dt);
        let y: number = this._controlPath.Calculate(u, dt);

        if ((this._direction == -1) && (this._y_n1 < y))
        {
            this._yMin = this._y_n1;
            this._direction = 1;
        }

        if ((this._direction == 1) && (this._y_n1 > y))
        {
            this._yMax = this._y_n1;
            this._direction = -1;
        }

        //this._y_n2 = this._y_n1;
        this._y_n1 = y;
        this._u = u;

        return y;
    }

    public TimeFunctionStepResponse(_t: number): number
    {
        return 0;
    }

    private _res: Complex = new Complex(0);
    private _tmp: Complex = new Complex(0);

    public TransferFunction(s: Complex): Complex
    {
        this._tmp.Copy(this._controller.TransferFunction(s));
        this._tmp.Mul(this._controlPath.TransferFunction(s));

        this._res.Copy(this._tmp);

        this._tmp.AddR(1);

        this._res.Div(this._tmp);

        return this._res;
    }

    public OptimizeGetNumParameters(): number
    {
        return this._controller.OptimizeGetNumParameters() + this._controlPath.OptimizeGetNumParameters();
    }

    public OptimizeSetParameters(params: number[], startIdx: number): void
    {
        this._controller.OptimizeSetParameters(params, startIdx);
        this._controlPath.OptimizeSetParameters(params, startIdx + this._controller.OptimizeGetNumParameters());
    }
}

//-------------------------------------------------------------------------------------------------
