
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { Complex } from "../../Framework/Math/Complex";

//-------------------------------------------------------------------------------------------------

export class ControlPID extends ControlElement
{
    private _Kp: number;
    private _Ki: number;
    private _Kd: number;

    private readonly _uMin: number;
    private readonly _uMax: number;

    private _e_n1: number = 0;
    private _e_sum: number = 0;

    public constructor(Kp: number, Ki: number, Kd: number, uMin: number, uMax: number)
    {
        super();

        this._Kp = Kp;
        this._Ki = Ki;
        this._Kd = Kd;

        this._uMin = uMin;
        this._uMax = uMax;
    }

    public GetType(): string
    {
        return "PID";
    }

    public GetLowestTimeConstant(): number
    {
        let tn: number = -1;
        let tv: number = -1;

        if (this._Ki != 0)
        {
            tn = this._Kp / this._Ki;
        }

        if (this._Kp != 0)
        {
            tv = this._Kd / this._Kp;
        }

        if (tn == -1)
            return tv;

        if (tv == -1)
            return tn;

        return Math.min(tn, tv);
    }

    public Initialize(_dt: number): void
    {
        this._e_n1 = 0;
        this._e_sum = 0;
    }

    public Calculate(e: number, dt: number): number
    {
        let Kp: number = this._Kp;
        let Ki: number = this._Ki;
        let Kd: number = this._Kd;
        let eSumOld: number = this._e_sum;

        this._e_sum += e;

        let u: number = Kp * e + Ki * dt * this._e_sum + Kd * (e - this._e_n1) / dt;

        this._e_n1 = e;

        if (u > this._uMax)
        {
            u = this._uMax;
            this._e_sum = eSumOld;
        }

        if (u < this._uMin)
        {
            u = this._uMin;
            this._e_sum = eSumOld;
        }

        return u;
    }

    public TimeFunctionStepResponse(t: number): number
    {
        let n: number = 10;
        let d: number = n * Math.exp(-Math.PI * n * n * t * t);
        return this._Kp + this._Ki * t + this._Kd * d;
    }

    private _res: Complex = new Complex(0);
    private _tmp: Complex = new Complex(0);

    public TransferFunction(s: Complex): Complex
    {
        // Kp + Ki / s + Kd * s

        this._tmp.Set(this._Ki, 0);
        this._tmp.Div(s);

        this._res.Copy(s);
        this._res.MulR(this._Kd);

        this._res.Add(this._tmp);
        this._res.AddR(this._Kp);

        return this._res;
    }

    public OptimizeGetNumParameters(): number
    {
        return 3;
    }

    public OptimizeSetParameters(params: number[], startIdx: number): void
    {
        this._Kp = params[startIdx + 0];
        this._Ki = params[startIdx + 1];
        this._Kd = params[startIdx + 2];

        if (this._Kp < 0)
            this._Kp = 0;

        if (this._Ki < 0)
            this._Ki = 0;

        if (this._Kd < 0)
            this._Kd = 0;
    }
}

//-------------------------------------------------------------------------------------------------
