
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { Complex } from "../../Framework/Math/Complex";

//-------------------------------------------------------------------------------------------------

export class ControlPath extends ControlElement
{
    private readonly _controlElementList: ControlElement[];

    public constructor(controlElementList?: ControlElement[])
    {
        super();

        if (controlElementList)
        {
            this._controlElementList = controlElementList;
        }
        else
        {
            this._controlElementList = [];
        }
    }

    public Add(controlelement: ControlElement): void
    {
        this._controlElementList.push(controlelement);
    }

    public GetType(): string
    {
        return "ControlPath";
    }

    public GetLowestTimeConstant(): number
    {
        let t: number = -1;
        let idx: number;

        for (idx = 0; idx < this._controlElementList.length; idx++)
        {
            let tt: number = this._controlElementList[idx].GetLowestTimeConstant();

            if (t == -1)
            {
                t = tt;
            }
            else
            {
                if ((tt != -1) && (tt < t))
                    t = tt;
            }
        }

        return t;
    }

    public Initialize(dt: number): void
    {
        let idx: number;

        for (idx = 0; idx < this._controlElementList.length; idx++)
        {
            this._controlElementList[idx].Initialize(dt);
        }
    }

    public Calculate(u: number, dt: number): number
    {
        let idx: number;
        let y: number = u;

        for (idx = 0; idx < this._controlElementList.length; idx++)
        {
            y = this._controlElementList[idx].Calculate(y, dt);
        }

        return y;
    }

    public TimeFunctionStepResponse(_t: number): number
    {
        return 0;
    }

    private _res: Complex = new Complex(0);

    public TransferFunction(s: Complex): Complex
    {
        if (this._controlElementList.length < 1)
        {
            this._res.Set(0, 0);
        }
        else
        {
            let idx: number;

            this._res.Copy(this._controlElementList[0].TransferFunction(s));

            for (idx = 1; idx < this._controlElementList.length; idx++)
                this._res.Mul(this._controlElementList[idx].TransferFunction(s));
        }

        return this._res;
    }

    public OptimizeGetNumParameters(): number
    {
        let numParams: number = 0;
        let idx: number;

        for (idx = 0; idx < this._controlElementList.length; idx++)
        {
            numParams += this._controlElementList[idx].OptimizeGetNumParameters();
        }

        return numParams;
    }

    public OptimizeSetParameters(params: number[], startIdx: number): void
    {
        let idx: number;
        let paramIdx: number = startIdx;

        for (idx = 0; idx < this._controlElementList.length; idx++)
        {
            this._controlElementList[idx].OptimizeSetParameters(params, paramIdx);
            paramIdx += this._controlElementList[idx].OptimizeGetNumParameters();
        }
    }
}

//-------------------------------------------------------------------------------------------------
