
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlPath } from "./ControlPath";
import { IDifferentialEvolutionRating } from "../../Framework/DifferentialEvolution/IDifferentialEvolutionRating";
import { DifferentialEvolution } from "../../Framework/DifferentialEvolution/DifferentialEvolution";

//-------------------------------------------------------------------------------------------------

export class ControlPathParamEstimation implements IDifferentialEvolutionRating
{
    //-------------------------------------------------------------------------------------------------

    private _controlPath: ControlPath;
    private _measData: [number, number][];

    private _stepHeight: number;
    private _stepTime: number;
    private _dt: number;

    //-------------------------------------------------------------------------------------------------

    public constructor(controlPath: ControlPath, measData: [number, number][])
    {
        this._controlPath = controlPath;
        this._measData = measData;

        let idx: number;

        this._dt = 0;

        for (idx = 1; idx < measData.length; idx++)
            this._dt += measData[idx][0] - measData[idx - 1][0];

        this._dt /= measData.length - 1;
    }

    //-------------------------------------------------------------------------------------------------

    public Optimize(limitMinParams: number, limitMaxParams: number, dither: boolean, maxNumSteps: number,
        stepHeight: number, stepTime: number): number[]
    {
        this._stepHeight = stepHeight;
        this._stepTime = stepTime;

        let numParams = this._controlPath.OptimizeGetNumParameters();

        let initMinParamsArr: number[] = [];
        let initMaxParamsArr: number[] = [];
        let limitMinParamsArr: number[] = [];
        let limitMaxParamsArr: number[] = [];
        let idx: number = 0;

        for (idx = 0; idx < numParams; idx++)
        {
            initMinParamsArr[idx] = 0;
            initMaxParamsArr[idx] = 1;
            limitMinParamsArr[idx] = limitMinParams;
            limitMaxParamsArr[idx] = limitMaxParams;
        }

        let diffEvo: DifferentialEvolution = new DifferentialEvolution(this, numParams, dither);
        let step: number = 0;

        diffEvo.CreateCandidates(initMinParamsArr, initMaxParamsArr, limitMinParamsArr, limitMaxParamsArr);

        for (step = 0; step < maxNumSteps; step++)
        {
            diffEvo.EvolutionStep();
        }

        return diffEvo.GetBestCandidate();
    }

    //-------------------------------------------------------------------------------------------------

    public RateCandidate(paramList: number[]): number
    {
        return this.Evaluate(paramList, this._stepHeight, this._stepTime);
    }

    //-------------------------------------------------------------------------------------------------

    private Evaluate(params: number[], stepHeight: number, stepTime: number): number
    {
        let dt: number = this._dt;
        let t: number = 0;
        let w: number = 0;
        let err: number = 0;
        let idx: number;

        this._controlPath.OptimizeSetParameters(params, 0);
        this._controlPath.Initialize(dt);

        for (idx = 0; idx < this._measData.length; idx++)
        {
            if (t >= stepTime)
                w = stepHeight;

            let y: number = this._controlPath.Calculate(w, dt);
            let dy: number = y - this._measData[idx][1];

            err += dy * dy;

            t += dt;
        }

        return err;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
