
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { ControlLoop } from "./ControlLoop";

//-------------------------------------------------------------------------------------------------

export class ControlSimulation
{
    //-------------------------------------------------------------------------------------------------

    private _controlElement: ControlElement;

    //-------------------------------------------------------------------------------------------------

    public constructor(controlElement: ControlElement)
    {
        this._controlElement = controlElement;
    }

    //-------------------------------------------------------------------------------------------------

    public StepResponseControlPath(stepHeight: number, stepTime: number, endTime: number): [[number, number][], [number, number][]]
    {
        let curveDt: number = endTime / 1000;
        let curveT: number = 0;
        let dt: number = this.GetDt(endTime);
        let t: number = 0;
        let u: number = 0;
        let curveU: [number, number][] = [];
        let curveY: [number, number][] = [];

        this._controlElement.Initialize(dt);

        while (t <= endTime)
        {
            if (t >= stepTime)
                u = stepHeight;

            let y: number = this._controlElement.Calculate(u, dt);

            curveT += dt;
            if (curveT >= curveDt)
            {
                curveU.push([t, u]);
                curveY.push([t, y]);

                curveT -= curveDt;
            }

            t += dt;
        }

        return [curveU, curveY];
    }

    //-------------------------------------------------------------------------------------------------

    public StepResponseControlLoop(stepHeight: number, stepTime: number, endTime: number): [[number, number][], [number, number][], [number, number][], number]
    {
        let curveDt: number = endTime / 1000;
        let curveT: number = 0;
        let dt: number = this.GetDt(endTime);
        let t: number = 0;
        let w: number = 0;
        let curveW: [number, number][] = [];
        let curveU: [number, number][] = [];
        let curveY: [number, number][] = [];
        let maxY: number = 0;

        this._controlElement.Initialize(dt);

        while (t <= endTime)
        {
            if (t >= stepTime)
                w = stepHeight;

            let y: number = this._controlElement.Calculate(w, dt);

            if ((w >= 0) && (y > maxY))
                maxY = y;

            if ((w < 0) && (y < maxY))
                maxY = y;

            curveT += dt;
            if (curveT >= curveDt)
            {
                let u: number = 0;

                if (this._controlElement instanceof ControlLoop)
                {
                    let ctrlLoop: ControlLoop = <ControlLoop>this._controlElement;
                    u = ctrlLoop.U;
                }

                curveW.push([t, w]);
                curveU.push([t, u]);
                curveY.push([t, y]);

                curveT -= curveDt;
            }

            t += dt;
        }

        return [curveW, curveU, curveY, maxY];
    }

    //-------------------------------------------------------------------------------------------------

    private GetDt(endTime: number): number
    {
        let MIN_NUM_POINTS: number = 100E3;
        let MAX_NUM_POINTS: number = 100E6;
        let lowT: number = this._controlElement.GetLowestTimeConstant();

        if (lowT == -1)
            return endTime / MIN_NUM_POINTS;

        let dt: number = lowT / 10000;
        let numPoints: number = endTime / dt;

        if (numPoints < MIN_NUM_POINTS)
            numPoints = MIN_NUM_POINTS;

        if (numPoints > MAX_NUM_POINTS)
            numPoints = MAX_NUM_POINTS;

        return endTime / numPoints;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
