
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Derivative } from "../../Framework/Math/Derivative";
import { MathEx } from "../../Framework/Math/MathEx";

//-------------------------------------------------------------------------------------------------

export class ControlTools
{
    //-------------------------------------------------------------------------------------------------

    public static CalcStepParams(uArr: [number, number][]): [number, number, number]
    {
        let stepHeight: number = 0;
        let endTime: number = 0;
        let stepTime: number = 0;
        let idx: number;

        for (idx = 0; idx < uArr.length; idx++)
        {
            if (uArr[idx][0] > endTime)
                endTime = uArr[idx][0];

            if (uArr[idx][1] > stepHeight)
                stepHeight = uArr[idx][1];
        }

        for (idx = 0; idx < uArr.length; idx++)
        {
            if (uArr[idx][1] > stepHeight / 2)
            {
                stepTime = uArr[idx][0];
                break;
            }
        }

        return [stepHeight, stepTime, endTime];
    }

    //-------------------------------------------------------------------------------------------------

    public static GetInflectionTangent(f: [number, number][], startX: number): [number, number]
    {
        const X: number = 0;
        const Y: number = 1;

        let d: [number, number][] = Derivative.DerivativeEx(f);
        let dd: [number, number][] = Derivative.DerivativeEx(d);
        let idx: number;

        for (idx = 1; idx < f.length; idx++)
        {
            if (f[idx][X] > startX)
                break;
        }

        for (; idx < f.length; idx++)
        {
            if (MathEx.Sign(dd[idx][Y]) != MathEx.Sign(dd[idx - 1][Y]))
                break;
        }

        if (idx == f.length)
            return [0, 0];

        let m: number = d[idx][Y];
        let n: number = f[idx][Y] - m * f[idx][X];

        return [m, n];
    }

    //-------------------------------------------------------------------------------------------------

    public static GetPidParamsZieglerNichols(tu: number, ku: number): [number, number, number]
    {
        let kp: number = 0.6 * ku;
        let tn: number = 0.5 * tu;
        let tv: number = 0.125 * tu;

        return [kp, kp / tn, kp * tv];
    }

    //-------------------------------------------------------------------------------------------------

    public static GetPidParamsLittleOvershot(tu: number, ku: number): [number, number, number]
    {
        let kp: number = 0.33 * ku;
        let tn: number = 0.5 * tu;
        let tv: number = 0.333 * tu;

        return [kp, kp / tn, kp * tv];
    }

    //-------------------------------------------------------------------------------------------------

    public static GetPidParamsNoOvershot(tu: number, ku: number): [number, number, number]
    {
        let kp: number = 0.2 * ku;
        let tn: number = 0.5 * tu;
        let tv: number = 0.333 * tu;

        return [kp, kp / tn, kp * tv];
    }

    //-------------------------------------------------------------------------------------------------

    public static ParseMeasDataString(measDataStr: string): [[number, number][], [number, number][]]
    {
        let uArr: [number, number][] = [];
        let yArr: [number, number][] = [];
        let measDataStrList: string[] = measDataStr.split("\n");
        let lineNum: number;

        for (lineNum = 0; lineNum < measDataStrList.length; lineNum++)
        {
            let line: string = measDataStrList[lineNum].trim();

            if (line.length > 0)
            {
                let valStrList: string[] = line.split(",");
                let t: number = Number(valStrList[0]);
                let u: number = Number(valStrList[1]);
                let y: number = Number(valStrList[2]);

                uArr.push([t, u]);
                yArr.push([t, y]);
            }
        }

        return [uArr, yArr];
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
