
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlLoop } from "./ControlLoop";
import { IControlOptimization } from "./IControlOptimization";
import { IDifferentialEvolutionRating } from "../../Framework/DifferentialEvolution/IDifferentialEvolutionRating";
import { DifferentialEvolution } from "../../Framework/DifferentialEvolution/DifferentialEvolution";
import { ControlOptimizationError } from "./ControlOptimizationError";

//-------------------------------------------------------------------------------------------------

export class ControllerOptimizationDifferentialEvolution implements IDifferentialEvolutionRating
{
    //-------------------------------------------------------------------------------------------------

    private readonly _NUM_TIME_STEPS: number = 5000;

    //-------------------------------------------------------------------------------------------------

    private _controlLoop: ControlLoop;
    private _controlOptimization: IControlOptimization;

    private _stepHeight: number;
    private _stepTime: number;
    private _endTime: number;
    //private _maxOvershot: number;
    private _calcError: ControlOptimizationError;

    //-------------------------------------------------------------------------------------------------

    public constructor(controlLoop: ControlLoop, controlOptimization: IControlOptimization)
    {
        this._controlLoop = controlLoop;
        this._controlOptimization = controlOptimization;
    }

    //-------------------------------------------------------------------------------------------------

    public Optimize(limitMinParams: number, limitMaxParams: number, dither: boolean, maxNumSteps: number,
        stepHeight: number, stepTime: number, endTime: number, maxOvershot: number,
        weightingCtrlOut: number, weightingOvershot: number): number[]
    {
        this._calcError = new ControlOptimizationError(this._controlLoop, stepHeight, maxOvershot, weightingCtrlOut, weightingOvershot);

        this._stepHeight = stepHeight;
        this._stepTime = stepTime;
        this._endTime = endTime;
        //this._maxOvershot = maxOvershot;

        let numParams = this._controlOptimization.OptimizeGetNumParameters();
        let diffEvo: DifferentialEvolution = new DifferentialEvolution(this, numParams, dither);
        let step: number = 0;

        let initMinParamsArr: number[] = [];
        let initMaxParamsArr: number[] = [];
        let limitMinParamsArr: number[] = [];
        let limitMaxParamsArr: number[] = [];
        let idx: number = 0;

        for (idx = 0; idx < numParams; idx++)
        {
            initMinParamsArr[idx] = 0;
            initMaxParamsArr[idx] = 1;
            limitMinParamsArr[idx] = limitMinParams;
            limitMaxParamsArr[idx] = limitMaxParams;
        }

        diffEvo.CreateCandidates(initMinParamsArr, initMaxParamsArr, limitMinParamsArr, limitMaxParamsArr);

        for (step = 0; step < maxNumSteps; step++)
        {
            diffEvo.EvolutionStep();
        }

        return diffEvo.GetBestCandidate();
    }

    //-------------------------------------------------------------------------------------------------

    public RateCandidate(paramList: number[]): number
    {
        return this.Evaluate(paramList, this._stepHeight, this._stepTime, this._endTime);
    }

    //-------------------------------------------------------------------------------------------------

    private Evaluate(params: number[], stepHeight: number, stepTime: number, endTime: number): number
    {
        let dt: number = endTime / this._NUM_TIME_STEPS;
        let t: number = 0;
        let w: number = 0;
        let err: number = 0;

        this._controlOptimization.OptimizeSetParameters(params, 0);
        this._controlLoop.Initialize(dt);

        while (t <= endTime)
        {
            if (t >= stepTime)
                w = stepHeight;

            let y: number = this._controlLoop.Calculate(w, dt);

            err += this._calcError.CalcError(w, y);

            t += dt;
        }

        return err;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
