
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlLoop } from "./ControlLoop";
import { IControlOptimization } from "./IControlOptimization";
import { ControlOptimizationError } from "./ControlOptimizationError";

//-------------------------------------------------------------------------------------------------

export class ControllerOptimizationLocalVariation
{
    //-------------------------------------------------------------------------------------------------

    private readonly _NUM_TIME_STEPS: number = 20000;

    //-------------------------------------------------------------------------------------------------

    private _controlLoop: ControlLoop;
    private _controlOptimization: IControlOptimization;
    private _calcError: ControlOptimizationError;

    //-------------------------------------------------------------------------------------------------

    public constructor(controlLoop: ControlLoop, controlOptimization: IControlOptimization)
    {
        this._controlLoop = controlLoop;
        this._controlOptimization = controlOptimization;
    }

    //-------------------------------------------------------------------------------------------------

    public Optimize(startParams: number[], startDelta: number[], maxNumSteps: number,
        stepHeight: number, stepTime: number, endTime: number, maxOvershot: number,
        weightingCtrlOut: number, weightingOvershot: number): number[]
    {
        this._calcError = new ControlOptimizationError(this._controlLoop, stepHeight, maxOvershot, weightingCtrlOut, weightingOvershot);

        let paramsCount: number = this._controlOptimization.OptimizeGetNumParameters();
        let params: number[] = [];
        let delta: number[] = [];
        let idx: number;

        for (idx = 0; idx < paramsCount; idx++)
        {
            params[idx] = startParams[idx];
            delta[idx] = startDelta[idx];
        }

        let step: number = 0;
        let err: number;

        err = this.Evaluate(params, stepHeight, stepTime, endTime);
        //console.log(err);

        while (step < maxNumSteps)
        {
            let better: boolean = false;

            for (idx = 0; idx < paramsCount; idx++)
            {
                let errAdd: number;
                let errSub: number;
                let p: number = params[idx];

                params[idx] = p + delta[idx];
                errAdd = this.Evaluate(params, stepHeight, stepTime, endTime);

                if (p > delta[idx])
                {
                    params[idx] = p - delta[idx];
                    errSub = this.Evaluate(params, stepHeight, stepTime, endTime);
                }
                else
                    errSub = err + 1;

                if ((errAdd < err) || (errSub < err))
                {
                    if (errAdd <= errSub)
                    {
                        err = errAdd;
                        params[idx] = p + delta[idx];
                    }
                    else
                    {
                        err = errSub;
                        params[idx] = p - delta[idx];
                    }

                    better = true;
                }
                else
                {
                    params[idx] = p;
                }
            }

            if (!better)
            {
                let smallEnough: boolean = true;

                for (idx = 0; idx < paramsCount; idx++)
                {
                    delta[idx] /= 2;

                    if (Math.abs(delta[idx]) > Math.abs(params[idx] / 1E6))
                        smallEnough = false;
                }

                if (smallEnough)
                    break;
            }

            step++;
        }

        return params;
    }

    //-------------------------------------------------------------------------------------------------

    private Evaluate(params: number[], stepHeight: number, stepTime: number, endTime: number): number
    {
        let dt: number = endTime / this._NUM_TIME_STEPS;
        let t: number = 0;
        let w: number = 0;
        let err: number = 0;

        this._controlOptimization.OptimizeSetParameters(params, 0);
        this._controlLoop.Initialize(dt);

        while (t <= endTime)
        {
            if (t >= stepTime)
                w = stepHeight;

            let y: number = this._controlLoop.Calculate(w, dt);

            err += this._calcError.CalcError(w, y);

            t += dt;
        }

        return err;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
