/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { ControlElement } from "./ControlElement";
import { ControlElementP } from "./ControlElementP";
import { ControlElementI } from "./ControlElementI";
import { ControlElementD } from "./ControlElementD";
import { ControlElementPT1 } from "./ControlElementPT1";
import { ControlElementPT2 } from "./ControlElementPT2";
import { ControlElementPD1 } from "./ControlElementPD1";
import { ControlElementDeadTime } from "./ControlElementDeadTime";

//---------------------------------------------------------------------------------------

declare var MathJax: any;

//-------------------------------------------------------------------------------------------------

export class GuiControlElement
{
    //-------------------------------------------------------------------------------------------------

    private readonly CTRL_TYPE_LIST: string[] = ["P", "I", "D", "PT1", "PT2", "PD1", "DeadTime"];

    private _ctrlType: string;
    private _idx: number;
    private _pic: HTMLObjectElement;

    private _hasParamK: boolean = false;
    private _hasParamT: boolean = false;
    private _hasParamD: boolean = false;
    private _htmlElemCtrl: HTMLElement;
    private _htmlElemParamK: HTMLInputElement;
    private _htmlElemParamT: HTMLInputElement;
    private _htmlElemParamD: HTMLInputElement;
    private _htmlElemTransferFunction: HTMLElement;

    private _k: number = 0;
    private _t: number = 0;
    private _d: number = 0;

    //-------------------------------------------------------------------------------------------------

    public get CtrlType(): string { return this._ctrlType; }
    public get HasParamK(): boolean { return this._hasParamK; }
    public get HasParamT(): boolean { return this._hasParamT; }
    public get HasParamD(): boolean { return this._hasParamD; }

    //-------------------------------------------------------------------------------------------------

    private constructor(pic: HTMLObjectElement, ctrlType: string, idx: number, hasParamK: boolean, hasParamT: boolean, hasParamD: boolean)
    {
        this._htmlElemTransferFunction = document.getElementById("LAPLACE_" + idx);

        this._pic = pic;
        this._ctrlType = ctrlType;
        this._idx = idx;

        this._hasParamK = hasParamK;
        this._hasParamT = hasParamT;
        this._hasParamD = hasParamD;

        if (ctrlType == "NONE")
            this._htmlElemCtrl = null;
        else
            this._htmlElemCtrl = pic.contentDocument.getElementById(ctrlType + "_CTRL_" + idx);

        this._htmlElemParamK = <HTMLInputElement>document.getElementById("PARAM_K_" + idx);
        this._htmlElemParamT = <HTMLInputElement>document.getElementById("PARAM_T_" + idx);
        this._htmlElemParamD = <HTMLInputElement>document.getElementById("PARAM_D_" + idx);
    }

    //-------------------------------------------------------------------------------------------------

    public static Create(ctrlType: string, idx: number, pic: HTMLObjectElement)
    {
        switch (ctrlType)
        {
            case "P": return new GuiControlElement(pic, "P", idx, true, false, false);
            case "I": return new GuiControlElement(pic, "I", idx, false, true, false);
            case "D": return new GuiControlElement(pic, "D", idx, true, false, false);
            case "PT1": return new GuiControlElement(pic, "PT1", idx, true, true, false);
            case "PT2": return new GuiControlElement(pic, "PT2", idx, true, true, true);
            case "PD1": return new GuiControlElement(pic, "PD1", idx, true, true, false);
            case "DeadTime": return new GuiControlElement(pic, "DeadTime", idx, false, true, false);
            case "NONE":
            default: return new GuiControlElement(pic, "NONE", idx, false, false, false);
        }
    }

    //-------------------------------------------------------------------------------------------------

    public CreateControl(): ControlElement
    {
        this.TakeParams();

        switch (this.CtrlType)
        {
            case "P": return new ControlElementP(this._k);
            case "I": return new ControlElementI(this._t);
            case "D": return new ControlElementD(this._k);
            case "PT1": return new ControlElementPT1(this._k, this._t);
            case "PT2": return new ControlElementPT2(this._k, this._t, this._d);
            case "PD1": return new ControlElementPD1(this._k, this._t);
            case "DeadTime": return new ControlElementDeadTime(this._t);
            default:
                throw new TypeError("CreateControl(): unknown type " + this.CtrlType);
        }
    }

    //-------------------------------------------------------------------------------------------------

    public Show(): void
    {
        let i: number;

        for (i = 0; i < this.CTRL_TYPE_LIST.length; i++)
        {
            let ctrlType: string = this.CTRL_TYPE_LIST[i];

            this._pic.contentDocument.getElementById(ctrlType + "_CTRL_" + this._idx).style.display = "none";
        }

        if (this._htmlElemCtrl != null)
            this._htmlElemCtrl.style.display = "initial";

        if (this._htmlElemTransferFunction != null)
        {
            this._htmlElemTransferFunction.textContent = "$$" + ControlElement.GetTransferFunctionLatex(this._ctrlType) + "$$";
            MathJax.Hub.Queue(["Typeset", MathJax.Hub, this._htmlElemTransferFunction]);
        }
    }

    //-------------------------------------------------------------------------------------------------

    public ShowParams(): void
    {
        this._htmlElemParamK.style.display = this.HasParamK ? "initial" : "none";
        this._htmlElemParamT.style.display = this.HasParamT ? "initial" : "none";
        this._htmlElemParamD.style.display = this.HasParamD ? "initial" : "none";
    }

    //-------------------------------------------------------------------------------------------------

    public TakeParams(): void
    {
        if (this.HasParamK)
        {
            if (this._htmlElemParamK.value.length == 0)
                throw new Error("row " + this._idx + " control element " + this.CtrlType + " invalid parameter K");

            this._k = Number(this._htmlElemParamK.value);
        }

        if (this.HasParamT)
        {
            if (this._htmlElemParamT.value.length == 0)
                throw new Error("row " + this._idx + " control element " + this.CtrlType + " invalid parameter T");

            this._t = Number(this._htmlElemParamT.value);
        }

        if (this.HasParamD)
        {
            if (this._htmlElemParamD.value.length == 0)
                throw new Error("row " + this._idx + " control element " + this.CtrlType + " invalid parameter D");

            this._d = Number(this._htmlElemParamD.value);
        }

        this.UpdateParamsInPic();
    }

    //-------------------------------------------------------------------------------------------------

    private UpdateParamsInPic(): void
    {
        if (this.HasParamK)
            this._pic.contentDocument.getElementById(this.CtrlType + "_CTRL_" + this._idx + "_K").textContent = "K = " + this._k.toPrecision(3);

        if (this.HasParamT)
            this._pic.contentDocument.getElementById(this.CtrlType + "_CTRL_" + this._idx + "_T").textContent = "T = " + this._t.toPrecision(3);

        if (this.HasParamD)
            this._pic.contentDocument.getElementById(this.CtrlType + "_CTRL_" + this._idx + "_D").textContent = "D = " + this._d.toPrecision(3);
    }

    //-------------------------------------------------------------------------------------------------

    private UpdateParamsInInputs(): void
    {
        if (this.HasParamK)
            this._htmlElemParamK.value = this._k.toPrecision(5);

        if (this.HasParamT)
            this._htmlElemParamT.value = this._t.toPrecision(5);

        if (this.HasParamD)
            this._htmlElemParamD.value = this._d.toPrecision(5);
    }

    //-------------------------------------------------------------------------------------------------

    public GetNumParams(): number
    {
        let cnt: number = 0;

        if (this.HasParamK)
            cnt++;

        if (this.HasParamT)
            cnt++;

        if (this.HasParamD)
            cnt++;

        return cnt;
    }

    //-------------------------------------------------------------------------------------------------

    public ShowParamVals(params: number[], paramIdx: number)
    {
        if (this.HasParamK)
            this._k = params[paramIdx++];

        if (this.HasParamT)
            this._t = params[paramIdx++];

        if (this.HasParamD)
            this._d = params[paramIdx++];

        this.UpdateParamsInPic();
        this.UpdateParamsInInputs();
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------

