
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { IDifferentialEvolutionRating } from "./IDifferentialEvolutionRating";

//-------------------------------------------------------------------------------------------------

class Candidate
{
    private _paramList: number[];
    public Rating: number;

    public get ParamList(): number[] { return this._paramList; }

    public constructor(numParams: number)
    {
        this._paramList = [];
        this._paramList.length = numParams;
        this.Rating = 0;
    }

    public Create(minParams: number[], maxParams: number[])
    {
        let idx: number;

        for (idx = 0; idx < this._paramList.length; idx++)
        {
            this._paramList[idx] = minParams[idx] + Math.random() * (maxParams[idx] - minParams[idx]);
        }
    }

    public MutateAndCrossover(candidateA: Candidate, candidateB: Candidate, candidateC: Candidate, weightingFactor: number, crossoverProbability: number,
        minParams: number[], maxParams: number[]): Candidate
    {
        let newCandidate: Candidate = new Candidate(this._paramList.length);
        let randomIdx: number = Math.floor(Math.random() * this._paramList.length);
        let idx: number;

        for (idx = 0; idx < this._paramList.length; idx++)
        {
            let r: number = Math.random();

            if ((idx == randomIdx) || (r < crossoverProbability))
            {
                let newParam: number = candidateC._paramList[idx] + weightingFactor * (candidateB._paramList[idx] - candidateA._paramList[idx]);

                if (newParam < minParams[idx])
                    newParam = minParams[idx];

                if (newParam > maxParams[idx])
                    newParam = maxParams[idx];

                newCandidate._paramList[idx] = newParam;
            }
            else
            {
                newCandidate._paramList[idx] = this._paramList[idx];
            }
        }

        return newCandidate;
    }
}

//-------------------------------------------------------------------------------------------------

export class DifferentialEvolution
{
    private readonly NUM_CANDIDATES_FACTOR: number = 12;

    private readonly WEIGHTING_FACTOR_MIN: number = 0.5;
    private readonly WEIGHTING_FACTOR_MAX: number = 1.0;
    private readonly WEIGHTING_FACTOR_FIX: number = 0.8;

    private readonly CROSSOVER_PROBABILITY: number = 0.8;

    private _dither: boolean;
    private _iDifferentialEvolutionRating: IDifferentialEvolutionRating;

    private _candidateList: Candidate[];
    private _tempCandidateList: Candidate[];

    private _limitMinParams: number[];
    private _limitMaxParams: number[];

    public constructor(iDifferentialEvolutionRating: IDifferentialEvolutionRating, numParams: number, dither?: boolean)
    {
        this._iDifferentialEvolutionRating = iDifferentialEvolutionRating;
        this._dither = (dither != null) ? dither : false;

        let numCandidates: number = numParams * this.NUM_CANDIDATES_FACTOR;

        this._candidateList = [];
        this._tempCandidateList = [];
        this._candidateList.length = numCandidates;
        this._tempCandidateList.length = numCandidates;

        let idx: number;

        for (idx = 0; idx < numCandidates; idx++)
        {
            this._candidateList[idx] = new Candidate(numParams);
        }
    }

    public GetBestCandidate(): number[]
    {
        let bestCandidate: Candidate = this._candidateList[0];
        let idx: number;

        for (idx = 1; idx < this._candidateList.length; idx++)
        {
            if (this._candidateList[idx].Rating < bestCandidate.Rating)
            {
                bestCandidate = this._candidateList[idx];
            }
        }

        return bestCandidate.ParamList;
    }

    public CreateCandidates(initMinParams: number[], initMaxParams: number[], limitMinParams: number[], limitMaxParams: number[])
    {
        this._limitMinParams = limitMinParams;
        this._limitMaxParams = limitMaxParams;

        let idx: number;

        for (idx = 0; idx < this._candidateList.length; idx++)
        {
            this._candidateList[idx].Create(initMinParams, initMaxParams);
            this._candidateList[idx].Rating = this._iDifferentialEvolutionRating.RateCandidate(this._candidateList[idx].ParamList);
        }
    }

    public EvolutionStep(): void
    {
        let idx: number;
        let candidateA: Candidate;
        let candidateB: Candidate;
        let candidateC: Candidate;
        let newCandidate: Candidate;
        let weightingFactor: number;

        if (this._dither)
            weightingFactor = this.WEIGHTING_FACTOR_MIN + Math.random() * (this.WEIGHTING_FACTOR_MAX - this.WEIGHTING_FACTOR_MIN);
        else
            weightingFactor = this.WEIGHTING_FACTOR_FIX;

        for (idx = 0; idx < this._candidateList.length; idx++)
        {
            [candidateA, candidateB, candidateC] = this.GetThreeRandomCandidates(idx);

            newCandidate = this._candidateList[idx].MutateAndCrossover(candidateA, candidateB, candidateC, weightingFactor, this.CROSSOVER_PROBABILITY,
                this._limitMinParams, this._limitMaxParams);

            newCandidate.Rating = this._iDifferentialEvolutionRating.RateCandidate(newCandidate.ParamList);

            if (newCandidate.Rating < this._candidateList[idx].Rating)
                this._tempCandidateList[idx] = newCandidate;
            else
                this._tempCandidateList[idx] = this._candidateList[idx];
        }

        for (idx = 0; idx < this._candidateList.length; idx++)
        {
            this._candidateList[idx] = this._tempCandidateList[idx];
        }
    }

    private GetRandomCandidateIndex(): number
    {
        return Math.floor(Math.random() * this._candidateList.length);
    }

    private GetThreeRandomCandidates(currentIdx: number): [Candidate, Candidate, Candidate]
    {
        let idxA: number;
        let idxB: number;
        let idxC: number;

        do
        {
            idxA = this.GetRandomCandidateIndex();
        }
        while (idxA == currentIdx);

        do
        {
            idxB = this.GetRandomCandidateIndex();
        }
        while ((idxB == currentIdx) || (idxB == idxA));

        do
        {
            idxC = this.GetRandomCandidateIndex();
        }
        while ((idxC == currentIdx) || (idxC == idxA) || (idxC == idxB));

        return [this._candidateList[idxA], this._candidateList[idxB], this._candidateList[idxC]];
    }
}

//-------------------------------------------------------------------------------------------------

