
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    ported from:

    numerical inverse laplace transform algorithm by
    Fausto Arinos de Almeida Barbuto (Calgary, Canada)
    E-mail: fausto_barbuto@yahoo.ca
    web: http://code.activestate.com/recipes/128243-numerical-inversion-of-laplace-transforms-using-th/

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "../Math/Complex";
import { IComplexFunction } from "../Math/IComplexFunction";
import { ILaplaceTransformInversion } from "./ILaplaceTransformInversion";

//-------------------------------------------------------------------------------------------------

/*
#################################################################
#   Function InvLap(t,omega,sigma,nint), numerically inverts a  #
#   Laplace transform F(s) into f(t) using the Fast Fourier     #
#   Transform (FFT) algorithm for a specific time "t", an       #
#   upper frequency limit "omega", a real parameter "sigma"     #
#   and the number of integration intervals "nint" .            #
#                                                               #
#   Function F(s) is defined in separate as Fs(s) (see code     #
#   below). Fs(s) has to be changed accordingly everytime the   #
#   user wants to invert a different function.                  #
#                                                               #
#   I suggest to use omega>100 and nint=50*omega. The higher    #
#   the values for omega, the more accurate the results will be #
#   in general, but at the expense of longer processing times.  #
#                                                               #
#   Sigma is a real number which must be a little bigger than   #
#   the real part of rightmost pole of the function F(s). For   #
#   example, F(s) = 1/s + 1/(s-2) + 1/(s+1) has poles for s=0,  #
#   s=2 and s=-1. Hence, sigma must be made equal to, say,      #
#   2.05 so as to keep all poles at the left of this value.     #
#   The analytical inverse for this simple function is          #
#   f(t) = 1 + exp(-t) + exp(2t). For t=1.25, omega=200,        #
#   nint=10000 and sigma=2.05, the numerical inversion yields   #
#   f(1.25) ~= 13.456844516, or -0.09% away from the actual     #
#   analytical result, 13.468998757 (results truncated to 9     #
#   decimal places). This is the example used in this code.     #
#                                                               #
#   Creator: Fausto Arinos de Almeida Barbuto (Calgary, Canada) #
#   Date: May 18, 2002                                          #
#   E-mail: fausto_barbuto@yahoo.ca                             #
#                                                               #
#   Reference:                                                  #
#   Huddleston, T. and Byrne, P: "Numerical Inversion of        #
#   Laplace Transforms", University of South Alabama, April     #
#   1999 (found at http://www.eng.usouthal.edu/huddleston/      #
#   SoftwareSupport/Download/Inversion99.doc)                   #
#                                                               #
#   Usage: invoke InvLap(t,omega,sigma,nint), for t>0.          #
#                                                               #
#################################################################
*/
export class LaplaceTransformInversionHuddleston implements ILaplaceTransformInversion
{
    private _sigma: number;
    private _omega: number;
    private _n: number;

    //-------------------------------------------------------------------------------------------------

    public GetName(): string { return "Huddleston"; }

    //-------------------------------------------------------------------------------------------------

    public constructor(sigma: number = 1, omega: number = 200, n: number = 10000)
    {
        this._sigma = sigma;
        this._omega = omega;
        this._n = n;
    }

    //-------------------------------------------------------------------------------------------------

    public Calculate(F: IComplexFunction, tArr: number[]): number[]
    {
        let yArr: number[] = [];
        let idx: number;
        let witi: Complex = new Complex(0);
        let wfti: Complex = new Complex(0);
        let fci: Complex = new Complex(0);
        let fcf: Complex = new Complex(0);

        for (idx = 0; idx < tArr.length; idx++)
        {
            let t: number = tArr[idx];

            if (t == 0)
            {
                yArr[idx] = 0;
                continue;
            }

            let sum: number = 0.0;
            let delta: number = this._omega / this._n;
            let wi: number = 0.0;
            let i: number;

            // The for-loop below computes the FFT Inversion Algorithm.
            // It is in fact the trapezoidal rule for numerical integration.

            for (i = 0; i < this._n; i++)
            {
                let wf: number = wi + delta;

                fci.Set(this._sigma, wi);
                fcf.Set(this._sigma, wf);

                witi.Set(0, wi * t);
                witi.Exp();
                witi.Mul(F.Function(fci));

                wfti.Set(0, wf * t);
                wfti.Exp();
                wfti.Mul(F.Function(fcf));

                sum = sum + 0.5 * (wf - wi) * (witi.Real + wfti.Real);
                wi = wf;
            }

            yArr[idx] = sum * Math.exp(this._sigma * t) / Math.PI;
        }

        return yArr;
    }

    //-------------------------------------------------------------------------------------------------

}

//-------------------------------------------------------------------------------------------------
