
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    ported from:

    numerical laplace transform inversion algorithm by
    Prof. Dr. Helmut Weber
    web: https://www.cs.hs-rm.de/~weber/

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "../Math/Complex";
import { IComplexFunction } from "../Math/IComplexFunction";
import { ILaplaceTransformInversion } from "./ILaplaceTransformInversion";
import { MathEx } from "../Math/MathEx";

//-------------------------------------------------------------------------------------------------

/*
    Funktion:   berechnet die Inverse f der Laplace-Transformierten

                          /00
                F(s) =   | exp(-s*t)f(t)dt, Re(s) > 0
                        0/

                mit dem Algorithmus von Koizumi.

    Parameter
    Eingabe:    n         - Zahl der beruechsichtigten Terme der unendlichen 
                            Reihe

    Literatur:  1. W. Squire, The Numerical Treatment of Laplace Transforms: 
                I. The Koizumi Inversion Method, International Journal for 
                Numerical Methods in Engineering, 20, 1697-1702 (1984)
                2. S. Koizumi, A new method of evaluation of the Heaviside 
                operational expression by Fourier sines, Phil. Mag. 19 (1935)
*/
export class LaplaceTransformInversionKoizumi implements ILaplaceTransformInversion
{
    private _n: number;
    private _coeff: number[];

    //-------------------------------------------------------------------------------------------------

    public GetName(): string { return "Koizumi"; }

    //-------------------------------------------------------------------------------------------------

    public constructor(n: number = 2048)
    {
        if (n < 1)
            throw new Error("n < 1");

        this._n = n;
        this._coeff = [];
        this._coeff.length = n;
    }

    //-------------------------------------------------------------------------------------------------

    public Calculate(F: IComplexFunction, tArr: number[]): number[]
    {
        let yArr: number[] = [];
        let T: number = MathEx.Max(tArr) * 4;
        let v1: number = Math.PI / (2 * T);
        let v2: number = 2.0 / T;
        let i: number;

        // Koeffizienten berechnen
        let z: Complex = new Complex(0);

        for (i = 0; i < this._n; i++)
        {
            z.Set(0, (1 - 2 * (i + 1)) * v1);
            this._coeff[i] = F.Function(z).Imag * v2;
        }

        // Fourierreihe fuer jedes t auswerten 
        for (i = 0; i < tArr.length; i++)
        {
            let t: number = tArr[i];

            if (t <= 0)
                yArr[i] = 0;
            else
            {
                let arg: number = t * v1;
                let ct: number = 2.0 * Math.cos(2.0 * arg);
                let c2: number = 0.0;
                let c1: number = this._coeff[this._n - 1];
                let k: number;

                for (k = this._n - 2; k >= 0; k--)
                {
                    let ck: number = ct * c1 - c2 + this._coeff[k];
                    c2 = c1;
                    c1 = ck;
                }

                yArr[i] = (c1 + c2) * Math.sin(arg);
            }
        }

        return yArr;
    }

    //-------------------------------------------------------------------------------------------------

}

//-------------------------------------------------------------------------------------------------
