
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "../Math/Complex";
import { IComplexFunction } from "../Math/IComplexFunction";
import { ILaplaceTransformInversion } from "./ILaplaceTransformInversion";
import { MathEx } from "../Math/MathEx";

//-------------------------------------------------------------------------------------------------

export class LaplaceTransformInversionStehfest implements ILaplaceTransformInversion
{
    private readonly _N: number = 16;

    private _coeff: number[];

    //-------------------------------------------------------------------------------------------------

    public GetName(): string { return "Stehfest"; }

    //-------------------------------------------------------------------------------------------------

    public constructor()
    {
        this._coeff = [];
        this._coeff.length = this._N;

        let n2: number = MathEx.Trunc(this._N / 2);
        let sign: number = 1;
        let i: number;

        if ((n2 % 2) != 0)
            sign = -1;

        for (i = 0; i < this._N; i++)
        {
            let kMin: number = MathEx.Trunc((i + 2) / 2);
            let kMax: number = i + 1;
            let k: number;

            if (kMax > n2)
                kMax = n2;

            this._coeff[i] = 0;
            sign = -sign;

            for (k = kMin; k <= kMax; k++)
            {
                this._coeff[i] += (Math.pow(k, n2) / MathEx.Factorial(k))
                    * (MathEx.Factorial(2 * k) / MathEx.Factorial(2 * k - i - 1))
                    / MathEx.Factorial(n2 - k) / MathEx.Factorial(k - 1) / MathEx.Factorial(i + 1 - k);
            }

            this._coeff[i] = sign * this._coeff[i];
        }
    }

    //-------------------------------------------------------------------------------------------------

    public Calculate(F: IComplexFunction, tArr: number[]): number[]
    {
        let yArr: number[] = [];
        let idx: number;
        let c: Complex = new Complex(0);

        for (idx = 0; idx < tArr.length; idx++)
        {
            let t: number = tArr[idx];

            if (t <= 0)
            {
                yArr[idx] = 0;
                continue;
            }

            let ln2t: number = Math.log(2.0) / t;
            let x: number = 0;
            let y: number = 0;
            let i: number;

            for (i = 0; i < this._coeff.length; i++)
            {
                x += ln2t;
                c.Set(x, 0);
                y += this._coeff[i] * F.Function(c).Real;
            }

            yArr[idx] = ln2t * y;
        }

        return yArr;
    }

    //-------------------------------------------------------------------------------------------------

}

//-------------------------------------------------------------------------------------------------
