﻿/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    ported from:

    numerical inverse laplace transform algorithm by
    Fernando Damian Nieuwveldt      
    email:fdnieuwveldt@gmail.com
    web: http://code.activestate.com/recipes/576934-numerical-inversion-of-the-laplace-transform-using/

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "../Math/Complex";
import { MathEx } from "../Math/MathEx";
import { IComplexFunction } from "../Math/IComplexFunction";
import { ILaplaceTransformInversion } from "./ILaplaceTransformInversion";

//-------------------------------------------------------------------------------------------------

/*
# Talbot suggested that the Bromwich line be deformed into a contour that begins
# and ends in the left half plane, i.e., z → −∞ at both ends.
# Due to the exponential factor the integrand decays rapidly
# on such a contour. In such situations the trapezoidal rule converge
# extraordinarily rapidly.
# For example here we compute the inverse transform of F(s) = 1/(s+1) at t = 1
#
# >>> error = Talbot(1,24)-exp(-1)
# >>> error
#   (3.3306690738754696e-015+0j)
#
# Talbot method is very powerful here we see an error of 3.3e-015
# with only 24 function evaluations
#
# Created by Fernando Damian Nieuwveldt      
# email:fdnieuwveldt@gmail.com
# Date : 25 October 2009
#
# Reference
# L.N.Trefethen, J.A.C.Weideman, and T.Schmelzer. Talbot quadratures
# and rational approximations. BIT. Numerical Mathematics,
# 46(3):653 670, 2006.
*/

export class LaplaceTransformInversionTalbot implements ILaplaceTransformInversion
{
    private _n: number;
    private _shift: number;

    //-------------------------------------------------------------------------------------------------

    public GetName(): string { return "Zakian"; }

    //-------------------------------------------------------------------------------------------------

    public constructor(n: number = 128, shift: number = 0.0)
    {
        this._n = n;
        this._shift = shift;
    }

    //-------------------------------------------------------------------------------------------------

    public Calculate(F: IComplexFunction, tArr: number[]): number[]
    {
        let yArr: number[] = [];
        let idx: number;
        let c1: number = 0.5017;
        let c2: number = 0.6407;
        let c3: number = 0.6122;
        let c4: Complex = new Complex(0, 0.2645);
        let z: Complex = new Complex(0);
        let dz: Complex = new Complex(0);
        let ans: Complex = new Complex(0);
        let e: Complex = new Complex(0);
        let r: Complex = new Complex(0);

        for (idx = 0; idx < tArr.length; idx++)
        {
            let t: number = tArr[idx];

            if (t <= 0)
            {
                yArr[idx] = 0;
                continue;
            }

            // Initiate the stepsize
            let h: number = 2 * Math.PI / this._n;

            // Shift contour to the right in case there is a pole on the positive real axis : Note the contour will
            // not be optimal since it was originally devoloped for function with
            // singularities on the negative real axis
            // For example take F(s) = 1/(s-1), it has a pole at s = 1, the contour needs to be shifted with one
            // unit, i.e shift  = 1. But in the test example no shifting is necessary
            // The for loop is evaluating the Laplace inversion at each point theta which is based on the trapezoidal   rule

            ans.Set(0);

            let k: number;
            for (k = 0; k <= this._n; k++)
            {
                let theta: number = -Math.PI + (k + 0.5) * h;
                let a: number = this._shift + this._n / t * (c1 * theta / Math.tan(c2 * theta) - c3);
                let b: number = this._n / t * (-c1 * c2 * theta / MathEx.Sqr(Math.sin(c2 * theta)) + c1 / Math.tan(c2 * theta));

                z.Copy(c4);
                z.MulR(this._n / t * theta);
                z.AddR(a);

                dz.Copy(c4);
                dz.MulR(this._n / t);
                dz.AddR(b);

                e.Copy(z);
                e.MulR(t);
                e.Exp();
                e.Mul(dz);
                e.Mul(F.Function(z));

                ans.Add(e);
            }

            r.Set(0, 2 * Math.PI);
            r.DivR(h);
            r.Inverse();
            r.Mul(ans);

            yArr[idx] = r.Real;
        }

        return yArr;
    }

    //-------------------------------------------------------------------------------------------------
}
