
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    ported from:

    numerical inverse laplace transform algorithm by
    Fausto Arinos de Almeida Barbuto (Calgary, Canada)
    E-mail: fausto_barbuto@yahoo.ca
    web: http://code.activestate.com/recipes/127469-numerical-inversion-of-laplace-transforms-through-/

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { Complex } from "../Math/Complex";
import { IComplexFunction } from "../Math/IComplexFunction";
import { ILaplaceTransformInversion } from "./ILaplaceTransformInversion";

//-------------------------------------------------------------------------------------------------

/*
#################################################################
#   Function Zakian(t), numerically inverts a Laplace transform #
#   F(s) into f(t) for a specific time "t" using the Zakian     #
#   algorithm.                                                  #
#                                                               #
#   Function F(s) is defined in separate as Fs(s) (see code     #
#   below). In the present case, F(s) = 1/(s-1), whose          #
#   analytical inversion is f(t) = exp(t), and t=1.0 thus       #
#   yielding f(1) = 2.71828180837 what is "e" with an error     #
#   of -7.4e-007%. Oscillatory systems may become innacurate    #
#   after the second cycle, though.                             #
#                                                               #
#   Fs(s) has to be changed accordingly everytime the user      #
#   wants to invert a different function.                       #
#                                                               #
#   Creator: Fausto Arinos de Almeida Barbuto (Calgary, Canada) #
#   Date: May 16, 2002                                          #
#   E-mail: fausto_barbuto@yahoo.ca                             #
#                                                               #
#   Reference:                                                  #
#   Huddleston, T. and Byrne, P: "Numerical Inversion of        #
#   Laplace Transforms", University of South Alabama, April     #
#   1999 (found at http://www.eng.usouthal.edu/huddleston/      #
#   SoftwareSupport/Download/Inversion99.doc)                   #
#                                                               #
#################################################################
*/
export class LaplaceTransformInversionZakian implements ILaplaceTransformInversion
{
    private _a: Complex[];
    private _k: Complex[];

    //-------------------------------------------------------------------------------------------------

    public GetName(): string { return "Zakian"; }

    //-------------------------------------------------------------------------------------------------

    public constructor()
    {
        this._a = [
            new Complex(12.83767675, 1.666063445),
            new Complex(12.22613209, 5.012718792),
            new Complex(10.93430308, 8.409673116),
            new Complex(8.776434715, 11.92185389),
            new Complex(5.225453361, 15.72952905)
        ];

        this._k = [
            new Complex(-36902.08210, 196990.4257),
            new Complex(61277.02524, -95408.62551),
            new Complex(-28916.56288, 18169.18531),
            new Complex(4655.361138, -1.901528642),
            new Complex(-118.7414011, -141.3036911)
        ];
    }

    //-------------------------------------------------------------------------------------------------

    public Calculate(F: IComplexFunction, tArr: number[]): number[]
    {
        let yArr: number[] = [];
        let idx: number;
        let c: Complex = new Complex();

        for (idx = 0; idx < tArr.length; idx++)
        {
            let t: number = tArr[idx];

            if (t <= 0)
            {
                yArr[idx] = 0;
                continue;
            }

            let sum: number = 0;
            let i: number;

            for (i = 0; i < 5; i++)
            {
                c.Copy(this._a[i]);
                c.DivR(t);

                c.Copy(F.Function(c));
                c.Mul(this._k[i]);

                sum += c.Real;
            }

            yArr[idx] = 2 * sum / t;
        }

        return yArr;
    }

    //-------------------------------------------------------------------------------------------------

}

//-------------------------------------------------------------------------------------------------
