
/*
    Copyright (C) 2017  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    based upon:

    Complex.js v2.0.1 11/02/2016
 
    Copyright (c) 2016, Robert Eisele (robert@xarg.org)
    Dual licensed under the MIT or GPL Version 2 licenses.
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

//-------------------------------------------------------------------------------------------------

export class Complex
{
    public static readonly ZERO: Complex = new Complex(0, 0);
    public static readonly ONE: Complex = new Complex(1, 0);
    public static readonly I: Complex = new Complex(0, 1);
    public static readonly PI: Complex = new Complex(Math.PI, 0);
    public static readonly E: Complex = new Complex(Math.E, 0);

    private static readonly EPSILON: number = 1e-16;

    private _re: number = 0;
    private _im: number = 0;

    public get Real(): number { return this._re; }
    public get Imag(): number { return this._im; }

    public constructor(re?: number, im?: number)
    {
        this.Set(re, im);
    }

    public static Create(a?: any, b?: any): Complex
    {
        let c: Complex = new Complex();
        c.Parse(a, b);
        return c;
    }

    public Set(re: number, im?: number)
    {
        this._re = (re != null) ? re : 0;
        this._im = (im != null) ? im : 0;
    }

    public Copy(src: Complex): void
    {
        this._re = src._re;
        this._im = src._im;
    }

    public clone(): any
    {
        return new Complex(this._re, this._im);
    }

    // Calculates the sign of a complex number, which is a normalized complex
    public Sign(): void
    {
        let abs: number = this.Abs();

        this._re /= abs;
        this._im /= abs;
    }

    // Adds two complex numbers
    public Add(a: Complex): void
    {
        this._re += a._re;
        this._im += a._im;
    }

    public AddR(a: number): void
    {
        this._re += a;
    }

    // Subtracts two complex numbers
    public Sub(a: Complex): void
    {
        this._re -= a._re;
        this._im -= a._im;
    }

    public SubR(a: number): void
    {
        this._re -= a;
    }

    // Multiplies two complex numbers
    public Mul(a: Complex): void
    {
        if ((a._im === 0) && (this._im === 0))
        {
            // Besides the addition/subtraction, this helps having a solution for rational Infinity
            this._re *= a._re;
            this._im = 0;
        }
        else
        {
            let re: number = this._re;
            this._re = this._re * a._re - this._im * a._im;
            this._im = re * a._im + this._im * a._re;
        }
    }

    public MulR(a: number): void
    {
        this._re *= a;
        this._im *= a;
    }

    // Divides two complex numbers
    public Div(a: Complex): void
    {
        let re: number = this._re;
        let im: number = this._im;

        if (a._im === 0)
        {
            if (a._re === 0)
            {
                // Divisor is zero
                this._re = (re !== 0) ? (re / 0) : 0;
                this._im = (im !== 0) ? (im / 0) : 0;
            }
            else
            {
                // Divisor is rational
                this._re /= a._re;
                this._im /= a._im;
            }
        }

        if (Math.abs(a._re) < Math.abs(a._im))
        {
            let x: number = a._re / a._im;
            let t: number = a._re * x + a._im;

            this._re = (re * x + im) / t;
            this._im = (im * x - re) / t;
        }
        else
        {
            let x: number = a._im / a._re;
            let t: number = a._im * x + a._re;

            this._re = (re + im * x) / t;
            this._im = (im - re * x) / t;
        }
    }

    public DivR(a: number): void
    {
        this._re /= a;
        this._im /= a;
    }

    // Calculate the power of two complex numbers
    public Pow(a: Complex): void
    {
        let re: number = this._re;
        let im: number = this._im;

        if ((re === 0) && (im === 0))
        {
            this._re = 0;
            this._im = 0;
            return;
        }

        let arg: number = Math.atan2(im, re);
        let loh: number = Complex.LogHypot(re, im);

        if (a._im === 0)
        {
            if ((im === 0) && (re >= 0))
            {
                this._re = Math.pow(re, a._re);
                this._im = 0;
                return;
            }
            else if (re === 0)
            {
                switch (a._re % 4)
                {
                    case 0:
                        this._re = Math.pow(im, a._re);
                        this._im = 0;
                        return;
                    case 1:
                        this._re = 0;
                        this._im = Math.pow(im, a._re);
                        return;
                    case 2:
                        this._re = -Math.pow(im, a._re);
                        this._im = 0;
                        return;
                    case 3:
                        this._re = 0;
                        this._im = -Math.pow(im, a._re);
                        return;
                }
            }
        }

        let x: number = Math.exp(a._re * loh - a._im * arg);
        let y: number = a._im * loh + a._re * arg;

        this._re = x * Math.cos(y);
        this._im = x * Math.sin(y);
    }

    // Calculate the complex square root
    public Sqrt(): void
    {
        let re: number = this._re;
        let im: number = this._im;
        let r: number = this.Abs();
        let rr: number;
        let ii: number;

        if (re >= 0)
        {
            if (im === 0)
            {
                this._re = Math.sqrt(re);
                this._im = 0;
                return;
            }

            rr = 0.5 * Math.sqrt(2.0 * (r + re));
        }
        else
        {
            rr = Math.abs(im) / Math.sqrt(2 * (r - re));
        }

        if (re <= 0)
            ii = 0.5 * Math.sqrt(2.0 * (r - re));
        else
            ii = Math.abs(im) / Math.sqrt(2 * (r + re));

        this._re = rr;
        this._im = (im < 0) ? -ii : ii;
    }

    // Calculate the complex exponent
    public Exp(): void
    {
        let a: number = Math.exp(this._re);

        this._re = a * Math.cos(this._im);
        this._im = a * Math.sin(this._im);
    }

    // Calculate the natural log
    public Log(): void
    {
        let re: number = this._re;
        let im: number = this._im;

        this._re = Complex.LogHypot(re, im);
        this._im = Math.atan2(im, re);
    }

    // Calculate the magniture of the complex number
    public Abs(): number
    {
        return Complex.Hypot(this._re, this._im);
    }

    // Calculate the angle of the complex number
    public Arg(): number
    {
        return Math.atan2(this._im, this._re);
    }

    // Calculate the sine of the complex number
    public Sin(): void
    {
        // sin(c) = (e^b - e^(-b)) / (2i)
        let re: number = this._re;
        let im: number = this._im;

        this._re = Math.sin(re) * Complex.CoshR(im);
        this._im = Math.cos(re) * Complex.SinhR(im);
    }

    // Calculate the cosine
    public Cos(): void
    {
        // cos(z) = (e^b + e^(-b)) / 2
        let re: number = this._re;
        let im: number = this._im;

        this._re = Math.cos(re) * Complex.CoshR(im);
        this._im = -Math.sin(re) * Complex.SinhR(im);
    }

    // Calculate the tangent
    public Tan()
    {
        // tan(c) = (e^(ci) - e^(-ci)) / (i(e^(ci) + e^(-ci)))

        let a: number = 2 * this._re;
        let b: number = 2 * this._im;
        let d: number = Math.cos(a) + Complex.CoshR(b);

        this._re = Math.sin(a) / d;
        this._im = Complex.SinhR(b) / d;
    }

    // Calculate the cotangent
    public Cot(): void
    {
        // cot(c) = i(e^(ci) + e^(-ci)) / (e^(ci) - e^(-ci))

        let a: number = 2 * this._re;
        let b: number = 2 * this._im;
        let d: number = Math.cos(a) - Complex.CoshR(b);

        this._re = -Math.sin(a) / d;
        this._im = Complex.SinhR(b) / d;
    }

    // Calculate the secant
    public Sec(): void
    {
        // sec(c) = 2 / (e^(ci) + e^(-ci))

        let a: number = this._re;
        let b: number = this._im;
        let d: number = 0.5 * Complex.CoshR(2 * b) + 0.5 * Math.cos(2 * a);

        this._re = Math.cos(a) * Complex.CoshR(b) / d;
        this._im = Math.sin(a) * Complex.SinhR(b) / d;
    }

    // Calculate the cosecans
    public Csc(): void
    {
        // csc(c) = 2i / (e^(ci) - e^(-ci))

        let a: number = this._re;
        let b: number = this._im;
        let d: number = 0.5 * Complex.CoshR(2 * b) - 0.5 * Math.cos(2 * a);

        this._re = Math.sin(a) * Complex.CoshR(b) / d;
        this._im = -Math.cos(a) * Complex.SinhR(b) / d;
    }

    // Calculate the complex arcus sinus
    public Asin(): void
    {
        // asin(c) = -i * log(ci + sqrt(1 - c^2))

        let a: number = this._re;
        let b: number = this._im;

        let t1: Complex = new Complex(b * b - a * a + 1, -2 * a * b);
        t1.Sqrt();

        let t2: Complex = new Complex(t1.Real - b, t1.Imag + a);
        t2.Log();

        this._re = t2.Imag;
        this._im = -t2.Real;
    }

    // Calculate the complex arcus cosinus
    public Acos(): void
    {
        // acos(c) = i * log(c - i * sqrt(1 - c^2))
        let a: number = this._re;
        let b: number = this._im;

        let t1: Complex = new Complex(b * b - a * a + 1, -2 * a * b);
        t1.Sqrt();

        let t2: Complex = new Complex(t1.Real - b, t1.Imag + a);
        t2.Log();

        this._re = Math.PI / 2 - t2.Imag;
        this._im = t2.Real;
    }

    // Calculate the complex arcus tangent
    public Atan(): void
    {
        // atan(c) = i / 2 log((i + x) / (i - x))
        let a: number = this._re;
        let b: number = this._im;

        if (a === 0)
        {
            if (b === 1)
            {
                this._re = 0;
                this._im = Infinity;
                return;
            }

            if (b === -1)
            {
                this._re = 0;
                this._im = -Infinity;
                return;
            }
        }

        let d: number = a * a + (1.0 - b) * (1.0 - b);
        let t1: Complex = new Complex((1 - b * b - a * a) / d, -2 * a / d);
        t1.Log();

        this._re = -0.5 * t1.Imag;
        this._im = 0.5 * t1.Real;
    }

    // Calculate the complex arcus cotangent
    public Acot(): void
    {
        // acot(c) = i / 2 log((c - i) / (c + i))
        let a: number = this._re;
        let b: number = this._im;

        if (b === 0)
        {
            this._re = Math.atan2(1, a);
            this._im = 0;
            return;
        }

        let d: number = a * a + b * b;
        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Atan();
        this.Copy(t);
    }

    // Calculate the complex arcus secant
    public Asec(): void
    {
        // asec(c) = -i * log(1 / c + sqrt(1 - i / c^2))
        let a: number = this._re;
        let b: number = this._im;

        if ((a === 0) && (b === 0))
        {
            this._re = 0;
            this._im = Infinity;
            return;
        }

        let d: number = a * a + b * b;

        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Acos();
        this.Copy(t);
    }

    // Calculate the complex arcus cosecans
    public Acsc(): void
    {
        // acsc(c) = -i * log(i / c + sqrt(1 - 1 / c^2))
        let a: number = this._re;
        let b: number = this._im;

        if ((a === 0) && (b === 0))
        {
            this._re = Math.PI / 2;
            this._im = Infinity;
            return;
        }

        let d: number = a * a + b * b;
        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Asin();
        this.Copy(t);
    }

    // Calculate the complex sinh
    public Sinh(): void
    {
        // sinh(c) = (e^c - e^-c) / 2
        let a: number = this._re;
        let b: number = this._im;

        this._re = Complex.SinhR(a) * Math.cos(b);
        this._im = Complex.CoshR(a) * Math.sin(b);
    }

    // Calculate the complex cosh
    public Cosh(): void
    {
        // cosh(c) = (e^c + e^-c) / 2
        let a: number = this._re;
        let b: number = this._im;

        this._re = Complex.CoshR(a) * Math.cos(b);
        this._im = Complex.SinhR(a) * Math.sin(b);
    }

    // Calculate the complex tanh
    public Tanh(): void
    {
        // tanh(c) = (e^c - e^-c) / (e^c + e^-c)
        let a: number = 2 * this._re;
        let b: number = 2 * this._im;
        let d: number = Complex.CoshR(a) + Math.cos(b);

        this._re = Complex.SinhR(a) / d;
        this._im = Math.sin(b) / d;
    }

    // Calculate the complex coth
    public Coth(): void
    {
        // coth(c) = (e^c + e^-c) / (e^c - e^-c)

        let a: number = 2 * this._re;
        let b: number = 2 * this._im;
        let d: number = Complex.CoshR(a) - Math.cos(b);

        this._re = Complex.SinhR(a) / d;
        this._im = -Math.sin(b) / d;
    }

    // Calculate the complex coth
    public Csch(): void
    {
        // csch(c) = 2 / (e^c - e^-c)
        let a: number = this._re;
        let b: number = this._im;
        let d: number = Math.cos(2 * b) - Complex.CoshR(2 * a);

        this._re = -2 * Complex.SinhR(a) * Math.cos(b) / d;
        this._im = 2 * Complex.CoshR(a) * Math.sin(b) / d;
    }

    // Calculate the complex sech
    public Sech(): void
    {
        // sech(c) = 2 / (e^c + e^-c)
        let a: number = this._re;
        let b: number = this._im;
        let d: number = Math.cos(2 * b) + Complex.CoshR(2 * a);

        this._re = 2 * Complex.CoshR(a) * Math.cos(b) / d;
        this._im = -2 * Complex.SinhR(a) * Math.sin(b) / d;
    }

    // Calculate the complex asinh
    public Asinh(): void
    {
        // asinh(c) = log(c + sqrt(c^2 + 1))
        let t: Complex = new Complex(this._im, -this._re);
        t.Asin();

        this._re = -t.Imag;
        this._im = t.Real;
    }

    // Calculate the complex asinh
    public Acosh(): void
    {
        // acosh(c) = log(c + sqrt(c^2 - 1))
        let t: Complex = new Complex();
        t.Copy(this);
        t.Acos();

        if (t.Imag <= 0)
        {
            this._re = -t.Imag;
            this._im = t.Real;
        }
        else
        {
            this._re = t.Imag;
            this._im = -t.Real;
        }
    }

    // Calculate the complex atanh
    public Atanh(): void
    {
        // atanh(c) = log((1+c) / (1-c)) / 2
        let a: number = this._re;
        let b: number = this._im;

        let noIM: boolean = (a > 1) && (b === 0);
        let oneMinus: number = 1 - a;
        let onePlus: number = 1 + a;
        let d: number = oneMinus * oneMinus + b * b;
        let x: Complex;

        if (d !== 0)
            x = new Complex((onePlus * oneMinus - b * b) / d, (b * oneMinus + onePlus * b) / d);
        else
            x = new Complex((a !== -1) ? (a / 0) : 0, (b !== 0) ? (b / 0) : 0);

        this._re = Complex.LogHypot(x.Real, x.Imag) / 2;
        this._im = Math.atan2(x.Imag, x.Real) / 2;

        if (noIM)
            this._im = -this._im;
    }

    // Calculate the complex acoth
    public Acoth(): void
    {
        // acoth(c) = log((c+1) / (c-1)) / 2
        let a: number = this._re;
        let b: number = this._im;

        if ((a === 0) && (b === 0))
        {
            this._re = 0;
            this._im = Math.PI / 2;
            return;
        }

        let d: number = a * a + b * b;
        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Atanh();
        this.Copy(t);
    }

    // Calculate the complex acsch
    public Acsch(): void
    {
        // acsch(c) = log((1+sqrt(1+c^2))/c)
        let a: number = this._re;
        let b: number = this._im;

        if (b === 0)
        {
            this._re = (a !== 0) ? Math.log(a + Math.sqrt(a * a + 1)) : Infinity;
            this._im = 0;
            return;
        }

        let d: number = a * a + b * b;
        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Asinh();
        this.Copy(t);
    }

    // Calculate the complex asech
    public Asech(): void
    {
        // asech(c) = log((1+sqrt(1-c^2))/c)
        let a: number = this._re;
        let b: number = this._im;

        if ((a === 0) && (b === 0))
        {
            this._re = Infinity;
            this._im = 0;
            return;
        }

        let d: number = a * a + b * b;
        let t: Complex;

        if (d !== 0)
            t = new Complex(a / d, -b / d);
        else
            t = new Complex((a !== 0) ? a / 0 : 0, (b !== 0) ? -b / 0 : 0);

        t.Acosh();
        this.Copy(t);
    }

    // Calculate the complex inverse 1/z
    public Inverse(): void
    {
        let a: number = this._re;
        let b: number = this._im;
        let d: number = a * a + b * b;

        this._re = a !== 0 ? a / d : 0;
        this._im = b !== 0 ? -b / d : 0;
    }

    // Returns the complex conjugate
    public Conjugate(): void
    {
        this._im = -this._im;
    }

    // Gets the negated complex number
    public Neg(): void
    {
        this._re = -this._re;
        this._im = -this._im;
    }

    // Ceils the actual complex number
    public Ceil(places: number): void
    {
        places = Math.pow(10, places);

        this._re = Math.ceil(this._re * places) / places;
        this._im = Math.ceil(this._im * places) / places;
    }

    // Floors the actual complex number
    public Floor(places: number): void
    {
        places = Math.pow(10, places);

        this._re = Math.floor(this._re * places) / places;
        this._im = Math.floor(this._im * places) / places;
    }

    // Rounds the actual complex number
    public Round(places: number): void
    {
        places = Math.pow(10, places);

        this._re = Math.round(this._re * places) / places;
        this._im = Math.round(this._im * places) / places;
    }

    // Compares two complex numbers
    public Equals(a: Complex): boolean
    {
        return (Math.abs(a._re - this._re) <= Complex.EPSILON) &&
            (Math.abs(a._im - this._im) <= Complex.EPSILON);
    }

    // Gets a string of the actual complex number
    public toString(): string
    {
        let a: number = this._re;
        let b: number = this._im;
        let ret: string = "";

        if (isNaN(a) || isNaN(b))
            return 'NaN';

        if (a !== 0)
            ret += a;

        if (b !== 0)
        {
            if (a !== 0)
                ret += (b < 0) ? " - " : " + ";
            else if (b < 0)
                ret += "-";

            b = Math.abs(b);

            if (1 !== b)
                ret += b;

            ret += "i";
        }

        if (ret.length === 0)
            return "0";

        return ret;
    }

    // Returns the actual number as a vector
    public toVector(): number[]
    {
        return [this._re, this._im];
    }

    // Returns the actual real value of the current object
    public valueOf(): number
    {
        if (this._im === 0)
            return this._re;

        return null;
    }

    // Checks if the given complex number is not a number
    public isNaN(): boolean
    {
        return isNaN(this._re) || isNaN(this._im);
    }

    private static CoshR(x: number): number
    {
        return (Math.exp(x) + Math.exp(-x)) * 0.5;
    }

    private static SinhR(x: number): number
    {
        return (Math.exp(x) - Math.exp(-x)) * 0.5;
    }

    private static Hypot(x: number, y: number): number
    {
        let a: number = Math.abs(x);
        let b: number = Math.abs(y);

        if (a < 3000 && b < 3000)
        {
            return Math.sqrt(a * a + b * b);
        }

        if (a < b)
        {
            a = b;
            b = x / y;
        }
        else
        {
            b = y / x;
        }

        return a * Math.sqrt(1 + b * b);
    }

    // Calculates log(sqrt(a^2+b^2)) in a way to avoid overflows
    private static LogHypot(a: number, b: number): number
    {
        let _a: number = Math.abs(a);
        let _b: number = Math.abs(b);

        if (a === 0)
            return Math.log(_b);

        if (b === 0)
            return Math.log(_a);

        if ((_a < 3000) && (_b < 3000))
            return Math.log(a * a + b * b) * 0.5;

        return Math.log(a / Math.cos(Math.atan2(b, a)));
    }

    public Parse(a: any, b: any): void
    {
        this._re = 0;
        this._im = 0;

        if (a instanceof Complex)
        {
            this.Copy(<Complex>a);
            return;
        }

        switch (typeof a)
        {
            case "string":
                this.ParseString(a);
                break;

            case "number":
                this._re = a;

                if (typeof b === "number")
                    this._im = b;

                break;

            case "object":

                if (("re" in a) && ("im" in a) && (typeof a.re === "number") && (typeof a.im === "number"))
                {
                    this._re = a["re"];
                    this._im = a["im"];
                }
                else if (("abs" in a) && ("arg" in a) && (typeof a["abs"] === "number") && (typeof a["arg"] === "number"))
                {
                    this._re = a["abs"] * Math.cos(a["arg"]);
                    this._im = a["abs"] * Math.sin(a["arg"]);
                }
                else if (("r" in a) && ("phi" in a) && (typeof a["r"] === "number") && (typeof a["phi"] === "number"))
                {
                    this._re = a["r"] * Math.cos(a["phi"]);
                    this._im = a["r"] * Math.sin(a["phi"]);
                }
                else
                    throw new SyntaxError("invalid parameter");

                break;

            default:
                throw new SyntaxError("invalid parameter");
        }
    }

    private ParseString(s: string): void
    {
        let tokens: RegExpMatchArray = s.match(/\d+\.?\d*e[+-]?\d+|\d+\.?\d*|\.\d+|./g);
        let plus: number = 1;
        let minus: number = 0;
        let i: number;

        if (tokens === null)
            throw new SyntaxError("invalid parameter");

        for (i = 0; i < tokens.length; i++)
        {
            let c: string = tokens[i];

            if (c === ' ' || c === '\t' || c === '\n')
            {
                /* void */
            }
            else if (c === '+')
                plus++;
            else if (c === '-')
                minus++;
            else if ((c === 'i') || (c === 'I'))
            {
                if (plus + minus === 0)
                {
                    throw new SyntaxError("invalid parameter");
                }

                if (tokens[i + 1] !== ' ' && !isNaN(Number(tokens[i + 1])))
                {
                    this._im += parseFloat((minus % 2 ? '-' : '') + tokens[i + 1]);
                    i++;
                }
                else
                {
                    this._im += parseFloat((minus % 2 ? '-' : '') + '1');
                }

                plus = minus = 0;

            }
            else
            {
                if (plus + minus === 0 || isNaN(Number(c)))
                    throw new SyntaxError("invalid parameter");

                if (tokens[i + 1] === 'i' || tokens[i + 1] === 'I')
                {
                    this._im += parseFloat((minus % 2 ? '-' : '') + c);
                    i++;
                }
                else
                    this._re += parseFloat((minus % 2 ? '-' : '') + c);

                plus = minus = 0;
            }
        }

        // Still something on the stack
        if (plus + minus > 0)
        {
            throw new SyntaxError("invalid parameter");
        }
    }
}

