
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

//-------------------------------------------------------------------------------------------------

export class Derivative
{

    //-------------------------------------------------------------------------------------------------

    public static Derivative(x: number[], y: number[]): number[]
    {
        if (x.length <= 1)
            throw new Error("Derivative: number of values must be greater than 1");

        if (x.length != y.length)
            throw new Error("Derivative: number of x and y values must be equal");

        let d: number[] = [];
        let idx: number;
        let sz: number = x.length - 1;

        d[0] = (y[1] - y[0]) / (x[1] - x[0]);
        d[sz] = (y[sz] - y[sz - 1]) / (x[sz] - x[sz - 1]);

        for (idx = 1; idx < sz; idx++)
        {
            let xm: number = (x[idx - 1] + x[idx] + x[idx + 1]) / 3.0;
            let ym: number = (y[idx - 1] + y[idx] + y[idx + 1]) / 3.0;
            let n: number;
            let z: number;

            n = (x[idx - 1] - xm) * (y[idx - 1] - ym);
            n += (x[idx] - xm) * (y[idx] - ym);
            n += (x[idx + 1] - xm) * (y[idx + 1] - ym);

            z = (x[idx - 1] - xm) * (x[idx - 1] - xm);
            z += (x[idx] - xm) * (x[idx] - xm);
            z += (x[idx + 1] - xm) * (x[idx + 1] - xm);

            if (z == 0)
                throw new Error("Derivative: error, some x values are equal");

            d[idx] = n / z;
        }

        return d;
    }

    //-------------------------------------------------------------------------------------------------

    public static DerivativeEx(f: [number, number][]): [number, number][]
    {
        const X: number = 0;
        const Y: number = 1;

        if (f.length <= 1)
            throw new Error("Derivative: number of values must be greater than 1");

        let d: [number, number][] = [];
        let idx: number;
        let sz: number = f.length - 1;

        d[0] = [f[0][X], (f[1][Y] - f[0][Y]) / (f[1][X] - f[0][X])];
        d[sz] = [f[sz][X], (f[sz][Y] - f[sz - 1][Y]) / (f[sz][X] - f[sz - 1][X])];

        for (idx = 1; idx < sz; idx++)
        {
            let xm: number = (f[idx - 1][X] + f[idx][X] + f[idx + 1][X]) / 3.0;
            let ym: number = (f[idx - 1][Y] + f[idx][Y] + f[idx + 1][Y]) / 3.0;
            let n: number;
            let z: number;

            n = (f[idx - 1][X] - xm) * (f[idx - 1][Y] - ym);
            n += (f[idx][X] - xm) * (f[idx][Y] - ym);
            n += (f[idx + 1][X] - xm) * (f[idx + 1][Y] - ym);

            z = (f[idx - 1][X] - xm) * (f[idx - 1][X] - xm);
            z += (f[idx][X] - xm) * (f[idx][X] - xm);
            z += (f[idx + 1][X] - xm) * (f[idx + 1][X] - xm);

            if (z == 0)
                throw new Error("Derivative: error, some x values are equal");

            d[idx] = [f[idx][X], n / z];
        }

        return d;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
