/*
    Copyright (C) 2018  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

//-------------------------------------------------------------------------------------------------

export class LinearEquationSystem
{
    //-------------------------------------------------------------------------------------------------

    /**
     * Löst das lineare Gleichungssystem: A * x = b
     * @param A Matrix A
     * @param b Vektor b
     * @param x Variablen x
     */
    public static GaussAlgorithm(A: number[][], b: number[], x: number[]): void
    {
        LinearEquationSystem.CheckArraySize(A, b);

        let cnt: number = b.length;

        x.length = cnt;

        // Pivotelement für Anfang suchen
        let pivotVal: number = 0;
        let pivotRow: number = -1;
        let pivotCol: number = -1;
        let row: number;
        let col: number;

        for (row = 0; row < cnt; row++)
        {
            for (col = 0; col < cnt; col++)
            {
                let val: number = Math.abs(A[row][col]);

                if (val > pivotVal)
                {
                    pivotVal = val;
                    pivotRow = row;
                    pivotCol = col;
                }
            }
        }

        // Arrays für Zeilen und Spalten Index erzeugen (für Pivotisierung)
        let rowIdxArr: number[] = [];
        let colIdxArr: number[] = [];

        rowIdxArr.length = cnt;
        colIdxArr.length = cnt;

        for (row = 0; row < cnt; row++)
        {
            rowIdxArr[row] = row;
            colIdxArr[row] = row;
        }

        // Gleichungssystem in Stufenform bringen
        let equation: number;

        for (equation = 0; equation < cnt - 1; equation++)
        {
            // Spalten und Zeilen Index tauschen
            if ((pivotRow < 0) || (pivotCol < 0))
                throw new Error("Matrix is zero");

            if (pivotRow != equation)
            {
                let temp: number = rowIdxArr[equation];
                rowIdxArr[equation] = rowIdxArr[pivotRow];
                rowIdxArr[pivotRow] = temp;
            }

            if (pivotCol != equation)
            {
                let temp: number = colIdxArr[equation];
                colIdxArr[equation] = colIdxArr[pivotCol];
                colIdxArr[pivotRow] = temp;
            }

            // Eliminierung, um Stufenform zu bekommen
            let equationRowIdx: number = rowIdxArr[equation];
            let equationColIdx: number = colIdxArr[equation];

            pivotVal = 0;
            pivotRow = -1;
            pivotCol = -1;

            for (row = equation + 1; row < cnt; row++)
            {
                let rowIdx: number = rowIdxArr[row];
                let f: number = A[rowIdx][equationColIdx] / A[equationRowIdx][equationColIdx];

                A[rowIdx][equationColIdx] = 0;

                for (col = equation + 1; col < cnt; col++)
                {
                    let colIdx: number = colIdxArr[col];
                    let val: number = A[rowIdx][colIdx] - f * A[equationRowIdx][colIdx];

                    A[rowIdx][colIdx] = val;

                    // nächstes Pivot Element suchen
                    val = Math.abs(val);
                    if (val > pivotVal)
                    {
                        pivotVal = val;
                        pivotRow = row;
                        pivotCol = col;
                    }
                }

                b[rowIdx] -= f * b[equationRowIdx];

                //LinearEquationSystem.Print(A, b, rowIdxArr, colIdxArr);
            }
        }

        // Variablen berechnen durch einsetzen
        for (equation = cnt - 1; equation >= 0; equation--)
        {
            let equationRowIdx: number = rowIdxArr[equation];
            let equationColIdx: number = colIdxArr[equation];
            let val: number = b[equationRowIdx];
            let col: number;

            for (col = equation + 1; col < cnt; col++)
            {
                let colIdx: number = colIdxArr[col];
                val -= A[equationRowIdx][colIdx] * x[colIdx];
            }

            x[equationColIdx] = val / A[equationRowIdx][equationColIdx];
        }
    }

    //-------------------------------------------------------------------------------------------------

    public static Print(A: number[][], b: number[], rowIdxArr: number[], colIdxArr: number[]): void
    {
        let cnt: number = b.length;
        let row: number;
        let col: number;

        for (row = 0; row < cnt; row++)
        {
            let rowIdx: number = rowIdxArr[row];
            let str: string = "";

            for (col = 0; col < cnt; col++)
            {
                let colIdx: number = colIdxArr[col];
                let val: number = A[rowIdx][colIdx];

                str += val.toPrecision(5) + "\t";
            }

            str += "\t=\t" + b[rowIdx].toPrecision(5);

            console.log(str);
        }
        console.log("------------------------------------")
    }

    //-------------------------------------------------------------------------------------------------

    private static CheckArraySize(A: number[][], b: number[]): void
    {
        if (b.length < 1)
            throw new RangeError("b.length < 1");

        if (A.length != b.length)
            throw new RangeError("A.length != b.length");

        if (A.length != A[0].length)
            throw new RangeError("A.length != A[0].length");
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
