
/*
    Copyright (C) 2016  Torsten Brischalle
    email: torsten@brischalle.de
    web: http://www.aaabbb.de
    
    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

import { CircuitType } from "./CircuitType";
import { ESeries } from "./ESeries";
import { ResistorResult } from "./ResistorResult";
import { ResistorCombinationResult } from "./ResistorCombinationResult";

//-------------------------------------------------------------------------------------------------

export class Resistor
{
    //-------------------------------------------------------------------------------------------------

    private _res: number;
    private _resSeries: number[];

    //-------------------------------------------------------------------------------------------------

    public get Res(): number { return this._res; }

    //-------------------------------------------------------------------------------------------------

    public constructor(res: number, resSeries: string)
    {
        this._res = res;
        this._resSeries = ESeries.GetSeries(resSeries);
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestFittingResistor(): ResistorResult
    {
        let exponent: number;
        let bestR: number = 0;
        let bestDiff: number = 1E21;

        for (exponent = -2; exponent <= 12; exponent++)
        {
            let factor: number = Math.pow(10, exponent);
            let rIdx: number;

            for (rIdx = 0; rIdx < this._resSeries.length; rIdx++)
            {
                let r: number = this._resSeries[rIdx] * factor;
                let diff: number = Math.abs(this._res - r) / this._res * 100;

                if (diff < bestDiff)
                {
                    bestDiff = diff;
                    bestR = r;
                }
            }
        }

        return new ResistorResult(bestR, bestDiff);
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestFittingCapacitor(): ResistorResult
    {
        let exponent: number;
        let bestR: number = 0;
        let bestDiff: number = 1E21;

        for (exponent = -12; exponent <= 2; exponent++)
        {
            let factor: number = Math.pow(10, exponent);
            let rIdx: number;

            for (rIdx = 0; rIdx < this._resSeries.length; rIdx++)
            {
                let r: number = this._resSeries[rIdx] * factor;
                let diff: number = Math.abs(this._res - r) / this._res * 100;

                if (diff < bestDiff)
                {
                    bestDiff = diff;
                    bestR = r;
                }
            }
        }

        return new ResistorResult(bestR, bestDiff);
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestSeriesResistor(): ResistorCombinationResult
    {
        let exponent1: number;
        let bestR1: number;
        let bestR2: number;
        let bestR: number = 0;
        let bestDiff: number = 1E21;

        for (exponent1 = -2; exponent1 <= 12; exponent1++)
        {
            let factor1: number = Math.pow(10, exponent1);
            let idx1: number;

            for (idx1 = 0; idx1 < this._resSeries.length; idx1++)
            {
                let r1: number = this._resSeries[idx1] * factor1;
                let exponent2: number;

                for (exponent2 = -2; exponent2 <= 12; exponent2++)
                {
                    let factor2: number = Math.pow(10, exponent2);
                    let idx2: number;

                    for (idx2 = 0; idx2 < this._resSeries.length; idx2++)
                    {
                        let r2: number = this._resSeries[idx2] * factor2;
                        let r: number = r1 + r2;
                        let diff: number = Math.abs(this._res - r) / this._res * 100;

                        if (diff < bestDiff)
                        {
                            bestDiff = diff;
                            bestR = r;
                            bestR1 = r1;
                            bestR2 = r2;
                        }
                    }
                }
            }
        }

        return new ResistorCombinationResult(bestR1, bestR2, bestR, bestDiff, CircuitType.SERIES);
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestParallelResistor(): ResistorCombinationResult
    {
        let exponent1: number;
        let bestR1: number;
        let bestR2: number;
        let bestR: number = 0;
        let bestDiff: number = 1E21;

        for (exponent1 = -2; exponent1 <= 12; exponent1++)
        {
            let factor1: number = Math.pow(10, exponent1);
            let idx1: number;

            for (idx1 = 0; idx1 < this._resSeries.length; idx1++)
            {
                let r1: number = this._resSeries[idx1] * factor1;
                let exponent2: number;

                for (exponent2 = -2; exponent2 <= 12; exponent2++)
                {
                    let factor2: number = Math.pow(10, exponent2);
                    let idx2: number;

                    for (idx2 = 0; idx2 < this._resSeries.length; idx2++)
                    {
                        let r2: number = this._resSeries[idx2] * factor2;
                        let r: number = r1 * r2 / (r1 + r2);
                        let diff: number = Math.abs(this._res - r) / this._res * 100;

                        if (diff < bestDiff)
                        {
                            bestDiff = diff;
                            bestR = r;
                            bestR1 = r1;
                            bestR2 = r2;
                        }
                    }
                }
            }
        }

        return new ResistorCombinationResult(bestR1, bestR2, bestR, bestDiff, CircuitType.PARALLEL);
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestFittingResistorCombination(): ResistorCombinationResult
    {
        let resCombList: Array<ResistorCombinationResult> = this.FindBestFittingResistorCombinationList(3);

        if (resCombList.length == 0)
            throw new Error("No resistor combination found!");

        return resCombList[0];
    }

    //-------------------------------------------------------------------------------------------------

    public FindBestFittingResistorCombinationList(maxNumResults: number): Array<ResistorCombinationResult>
    {
        let exponent: number;
        let resultList: Array<ResistorCombinationResult> = new Array<ResistorCombinationResult>();

        loop:
        for (exponent = -2; exponent <= 12; exponent++)
        {
            let factor: number = Math.pow(10, exponent);
            let rIdx: number;

            for (rIdx = 0; rIdx < this._resSeries.length; rIdx++)
            {
                let r1: number = this._resSeries[rIdx] * factor;

                if (Math.round(r1 * 1E7) == Math.round(this._res * 1E7))
                {
                    resultList = new Array<ResistorCombinationResult>();
                    resultList.push(new ResistorCombinationResult(r1, 0, r1, 0, CircuitType.SINGLE));
                    break loop;
                }

                let r2Calc: number;
                let r: number;
                let diff: number;
                let result: ResistorCombinationResult;

                // series
                r2Calc = this._res - r1;
                if (r2Calc > 0)
                {
                    let r2: [number, number] = this.FindNearestResistors(r2Calc);

                    r = r1 + r2[0];
                    diff = Math.abs(this._res - r) / this._res * 100;
                    result = new ResistorCombinationResult(r1, r2[0], r, diff, CircuitType.SERIES);
                    this.InsertResultInSortedArray(resultList, result, maxNumResults);

                    r = r1 + r2[1];
                    diff = Math.abs(this._res - r) / this._res * 100;
                    result = new ResistorCombinationResult(r1, r2[1], r, diff, CircuitType.SERIES);
                    this.InsertResultInSortedArray(resultList, result, maxNumResults);
                }

                // parallel
                r2Calc = this._res * r1 / (r1 - this._res);
                if (r2Calc > 0)
                {
                    let r2: [number, number] = this.FindNearestResistors(r2Calc);

                    r = r1 * r2[0] / (r1 + r2[0]);
                    diff = Math.abs(this._res - r) / this._res * 100;
                    result = new ResistorCombinationResult(r1, r2[0], r, diff, CircuitType.PARALLEL);
                    this.InsertResultInSortedArray(resultList, result, maxNumResults);

                    r = r1 * r2[1] / (r1 + r2[1]);
                    diff = Math.abs(this._res - r) / this._res * 100;
                    result = new ResistorCombinationResult(r1, r2[1], r, diff, CircuitType.PARALLEL);
                    this.InsertResultInSortedArray(resultList, result, maxNumResults);
                }
            }

        }

        return resultList;
    }

    //-------------------------------------------------------------------------------------------------

    private FindNearestResistors(res: number): [number, number]
    {
        let exponent: number = Math.floor(Math.log(res) / Math.log(10));
        let resLo: number = 0;
        let resHi: number = 0;
        let idx: number;
        let expo: number;

        loopHi:
        for (expo = exponent; expo <= 12; expo++)
        {
            let factor: number = Math.pow(10, exponent);

            for (idx = 0; idx < this._resSeries.length; idx++)
            {
                let r: number = this._resSeries[idx] * factor;

                if (r >= res)
                {
                    resHi = r;
                    break loopHi;
                }
            }
        }

        loopLo:
        for (expo = exponent; expo >= -2; expo--)
        {
            let factor: number = Math.pow(10, exponent);

            for (idx = this._resSeries.length - 1; idx >= 0; idx--)
            {
                let r: number = this._resSeries[idx] * factor;

                if (r <= res)
                {
                    resLo = r;
                    break loopLo;
                }
            }
        }

        return [resLo, resHi];
    }

    //-------------------------------------------------------------------------------------------------

    private InsertResultInSortedArray(resultList: Array<ResistorCombinationResult>, result: ResistorCombinationResult, maxNumResults: number)
    {
        let idx: number;

        for (idx = 0; idx < resultList.length; idx++)
        {
            if ((result.R1 == resultList[idx].R1) && (result.R2 == resultList[idx].R2))
                return;

            if ((result.R1 == resultList[idx].R2) && (result.R2 == resultList[idx].R1))
                return;
        }

        for (idx = 0; idx < resultList.length; idx++)
        {
            if (result.Diff < resultList[idx].Diff)
                break;
        }

        resultList.splice(idx, 0, result);

        if (resultList.length > maxNumResults)
            resultList.length = maxNumResults;
    }

    //-------------------------------------------------------------------------------------------------

    public GetNextResistors(count: number): number[]
    {
        let resList: number[] = [];
        let exponent: number = Math.ceil(Math.log(this._res) / Math.log(10) - 1E-12) - 2;

        while (resList.length < count)
        {
            let idx: number;

            for (idx = 0; idx < this._resSeries.length; idx++)
            {
                let r: number = this._resSeries[idx] * Math.pow(10, exponent);

                if (r >= this._res)
                {
                    resList.push(r);

                    if (resList.length >= count)
                        break;
                }
            }

            exponent++;
        }

        return resList;
    }

    //-------------------------------------------------------------------------------------------------

    public GetPrevResistors(count: number): number[]
    {
        let resList: number[] = [];
        let exponent: number = Math.ceil(Math.log(this._res) / Math.log(10) - 1E-12) + 2;

        while (resList.length < count)
        {
            let idx: number;

            for (idx = this._resSeries.length - 1; idx >= 0; idx--)
            {
                let r: number = this._resSeries[idx] * Math.pow(10, exponent);

                if (r <= this._res)
                {
                    resList.push(r);

                    if (resList.length >= count)
                        break;
                }
            }

            exponent--;
        }

        return resList;
    }

    //-------------------------------------------------------------------------------------------------
}

//-------------------------------------------------------------------------------------------------
